import time
import warnings
from abc import ABC
from collections.abc import Generator, Iterator, Sequence
from contextlib import contextmanager
from typing import cast
from unittest import mock

from dagster import AssetCheckKey, AssetKey, BetaWarning, materialize
from dagster._core.definitions.assets.graph.remote_asset_graph import RemoteAssetNode
from dagster._core.definitions.assets.job.asset_job import IMPLICIT_ASSET_JOB_NAME
from dagster._core.definitions.selector import JobSelector, ScheduleSelector, SensorSelector
from dagster._core.execution.backfill import BulkActionStatus, PartitionBackfill
from dagster._core.remote_representation.external import (
    CompoundID,
    RemoteJob,
    RemoteSchedule,
    RemoteSensor,
)
from dagster._core.utils import make_new_backfill_id
from dagster._core.workspace.context import RemoteDefinition, WorkspaceRequestContext
from dagster._core.workspace.permissions import Permissions
from dagster_graphql.client.query import (
    LAUNCH_PARTITION_BACKFILL_MUTATION,
    LAUNCH_PIPELINE_EXECUTION_MUTATION,
    LAUNCH_PIPELINE_REEXECUTION_MUTATION,
)
from dagster_graphql.test.utils import (
    GqlAssetKey,
    execute_dagster_graphql,
    infer_job_selector,
    infer_schedule_selector,
    infer_sensor_selector,
)

from dagster_graphql_tests.graphql.graphql_context_test_suite import (
    GraphQLContextVariant,
    make_graphql_context_test_suite,
)
from dagster_graphql_tests.graphql.repo import owned_asset, unowned_asset
from dagster_graphql_tests.graphql.test_assets import WIPE_ASSETS
from dagster_graphql_tests.graphql.test_partition_backfill import (
    CANCEL_BACKFILL_MUTATION,
    GET_PARTITION_BACKFILLS_QUERY,
)
from dagster_graphql_tests.graphql.test_runs import DELETE_RUN_MUTATION
from dagster_graphql_tests.graphql.test_scheduler import (
    GET_SCHEDULE_STATE_QUERY,
    START_SCHEDULES_QUERY,
    STOP_SCHEDULES_QUERY,
)
from dagster_graphql_tests.graphql.test_sensors import GET_SENSOR_STATUS_QUERY, START_SENSORS_QUERY

warnings.filterwarnings("ignore", category=BetaWarning, message=r"Parameter `owners` .*")

GET_JOB_PERMISSIONS_QUERY = """
query JobPermissionsQuery($selector: PipelineSelector!) {
    pipelineOrError(params: $selector) {
        __typename
        ... on Pipeline {
            hasLaunchExecutionPermission
            hasLaunchReexecutionPermission
        }
    }
}
"""

GET_ASSET_PERMISSIONS_QUERY = """
query AssetPermissionsQuery($assetKey: AssetKeyInput!) {
    assetNodeOrError(assetKey: $assetKey) {
        __typename
        ... on AssetNode {
            hasWipePermission
        }
    }
}
"""

BaseTestSuite = make_graphql_context_test_suite(
    # no need to test all storages... just pick one that supports launching
    context_variants=[
        GraphQLContextVariant.sqlite_with_default_run_launcher_managed_grpc_env(
            None, "test_location"
        ),
    ]
)


class BaseDefinitionOwnerPermissionsTestSuite(ABC):
    def get_owned_definitions(self) -> set[str | AssetKey]:
        return set(
            [
                AssetKey(["owned_asset"]),
                AssetKey(["owned_partitioned_asset"]),
                "owned_job",
                "owned_partitioned_job",
                "owned_schedule",
                "owned_sensor",
            ]
        )

    def get_unowned_definitions(self) -> set[str | AssetKey]:
        return set(
            [
                AssetKey(["unowned_asset"]),
                AssetKey(["unowned_partitioned_asset"]),
                "unowned_job",
                "unowned_partitioned_job",
                "unowned_schedule",
                "unowned_sensor",
            ]
        )

    def _get_selector(
        self, context, def_type: type[RemoteDefinition], def_handle: str | AssetKey
    ) -> AssetKey | JobSelector | ScheduleSelector | SensorSelector:
        if isinstance(def_handle, AssetKey):
            return def_handle

        assert type(def_handle) is str
        if def_type is RemoteJob:
            return JobSelector.from_graphql_input(infer_job_selector(context, def_handle))
        if def_type is RemoteSensor:
            return SensorSelector.from_graphql_input(infer_sensor_selector(context, def_handle))
        if def_type is RemoteSchedule:
            return ScheduleSelector.from_graphql_input(infer_schedule_selector(context, def_handle))
        raise Exception(f"Unknown definition type {def_type}")

    def get_partition_set_by_job_name(self, context, job_name: str):
        job_selector = JobSelector.from_graphql_input(infer_job_selector(context, job_name))
        repository = context.get_code_location(job_selector.location_name).get_repository(
            job_selector.repository_name
        )
        return next(
            iter(ps for ps in repository.get_partition_sets() if ps.job_name == job_name), None
        )

    def graphql_launch_asset_backfill(self, context, asset_keys: list[AssetKey]):
        result = execute_dagster_graphql(
            context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "partitionNames": ["a", "b", "c"],
                    "assetSelection": [key.to_graphql_input() for key in asset_keys],
                }
            },
        )
        assert result.data
        typename = result.data["launchPartitionBackfill"]["__typename"]
        backfill_id = (
            result.data["launchPartitionBackfill"].get("backfillId")
            if typename == "LaunchBackfillSuccess"
            else None
        )
        return typename, backfill_id

    def graphql_get_job_permissions(
        self, context, job_name: str, asset_selection: Sequence[AssetKey] | None = None
    ):
        gql_asset_selection = (
            cast("Sequence[GqlAssetKey]", [key.to_graphql_input() for key in asset_selection])
            if asset_selection
            else None
        )

        selector = infer_job_selector(context, job_name, asset_selection=gql_asset_selection)
        result = execute_dagster_graphql(
            context,
            GET_JOB_PERMISSIONS_QUERY,
            variables={"selector": selector},
        )
        assert result.data
        return (
            result.data["pipelineOrError"]["hasLaunchExecutionPermission"],
            result.data["pipelineOrError"]["hasLaunchReexecutionPermission"],
        )

    def graphql_launch_job_backfill(self, context, job_name: str):
        partition_set = self.get_partition_set_by_job_name(context, job_name)
        assert partition_set is not None

        result = execute_dagster_graphql(
            context,
            LAUNCH_PARTITION_BACKFILL_MUTATION,
            variables={
                "backfillParams": {
                    "selector": {
                        "repositorySelector": {
                            "repositoryLocationName": partition_set.repository_handle.location_name,
                            "repositoryName": partition_set.repository_handle.repository_name,
                        },
                        "partitionSetName": partition_set.name,
                    },
                    "partitionNames": ["a", "b", "c"],
                }
            },
        )
        assert result.data
        typename = result.data["launchPartitionBackfill"]["__typename"]
        backfill_id = (
            result.data["launchPartitionBackfill"].get("backfillId")
            if typename == "LaunchBackfillSuccess"
            else None
        )
        return typename, backfill_id

    def graphql_has_job_backfill_permissions(self, context, job_name: str):
        partition_set = self.get_partition_set_by_job_name(context, job_name)
        assert partition_set is not None

        result = execute_dagster_graphql(
            context,
            GET_PARTITION_BACKFILLS_QUERY,
            variables={
                "repositorySelector": {
                    "repositoryLocationName": partition_set.repository_handle.location_name,
                    "repositoryName": partition_set.repository_handle.repository_name,
                },
                "partitionSetName": partition_set.name,
            },
        )
        partition_set = result.data["partitionSetOrError"]
        return partition_set["hasLaunchBackfillPermission"], partition_set[
            "hasCancelBackfillPermission"
        ]

    def graphql_cancel_backfill(self, context, backfill_id: str):
        result = execute_dagster_graphql(
            context,
            CANCEL_BACKFILL_MUTATION,
            variables={"backfillId": backfill_id},
        )
        assert result.data
        return result.data["cancelPartitionBackfill"]["__typename"]

    def graphql_launch_asset_run(self, context, asset_selection: list[AssetKey]):
        gql_asset_selection = (
            cast("Sequence[GqlAssetKey]", [key.to_graphql_input() for key in asset_selection])
            if asset_selection
            else None
        )
        selector = infer_job_selector(
            context,
            IMPLICIT_ASSET_JOB_NAME,
            asset_selection=gql_asset_selection,
        )
        result = execute_dagster_graphql(
            context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "mode": "default",
                }
            },
        )
        assert result.data
        return result.data["launchPipelineExecution"]["__typename"]

    def graphql_launch_job_run(self, context, job_name: str):
        selector = infer_job_selector(context, job_name)

        result = execute_dagster_graphql(
            context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "mode": "default",
                }
            },
        )
        assert result.data
        typename = result.data["launchPipelineExecution"]["__typename"]
        run_id = (
            result.data["launchPipelineExecution"]["run"]["runId"]
            if typename == "LaunchRunSuccess"
            else None
        )
        return typename, run_id

    def graphql_launch_asset_job(
        self, context, asset_selection: list[AssetKey], asset_check_selection: list[AssetCheckKey]
    ):
        selector = infer_job_selector(
            context,
            IMPLICIT_ASSET_JOB_NAME,
            asset_selection=[{"path": asset_key.path} for asset_key in asset_selection],
            asset_check_selection=[
                {"assetKey": {"path": check_key.asset_key.path}, "name": check_key.name}
                for check_key in asset_check_selection
            ],
        )
        result = execute_dagster_graphql(
            context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={"executionParams": {"selector": selector, "mode": "default"}},
        )
        assert result.data
        return result.data["launchPipelineExecution"]["__typename"]

    def graphql_launch_job_reexecution(
        self,
        context,
        job_name: str,
        run_id: str,
        asset_selection: list[AssetKey] | None = None,
        asset_check_selection: list[AssetCheckKey] | None = None,
    ):
        selector = infer_job_selector(
            context,
            job_name,
            asset_selection=[{"path": key.path} for key in asset_selection]
            if asset_selection
            else None,
            asset_check_selection=[
                {"assetKey": {"path": check_key.asset_key.path}, "name": check_key.name}
                for check_key in asset_check_selection
            ]
            if asset_check_selection
            else None,
        )

        result = execute_dagster_graphql(
            context,
            LAUNCH_PIPELINE_REEXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "executionMetadata": {
                        "rootRunId": run_id,
                        "parentRunId": run_id,
                    },
                    "mode": "default",
                }
            },
        )
        assert result.data
        return result.data["launchPipelineReexecution"]["__typename"]

    def graphql_start_schedule(self, context, schedule_name: str):
        selector = infer_schedule_selector(context, schedule_name)

        result = execute_dagster_graphql(
            context,
            START_SCHEDULES_QUERY,
            variables={
                "scheduleSelector": selector,
            },
        )
        assert result.data
        typename = result.data["startSchedule"]["__typename"]
        schedule_id = (
            result.data["startSchedule"]["scheduleState"]["id"]
            if typename == "ScheduleStateResult"
            else None
        )
        return typename, schedule_id

    def graphql_get_schedule_permissions(self, context, schedule_name: str):
        selector = infer_schedule_selector(context, schedule_name)
        result = execute_dagster_graphql(
            context,
            GET_SCHEDULE_STATE_QUERY,
            variables={
                "scheduleSelector": selector,
            },
        )
        assert result.data
        return (
            result.data["scheduleOrError"]["scheduleState"]["hasStartPermission"],
            result.data["scheduleOrError"]["scheduleState"]["hasStopPermission"],
        )

    def graphql_get_sensor_permissions(self, context, sensor_name: str):
        selector = infer_sensor_selector(context, sensor_name)
        result = execute_dagster_graphql(
            context,
            GET_SENSOR_STATUS_QUERY,
            variables={
                "sensorSelector": selector,
            },
        )
        assert result.data
        return (
            result.data["sensorOrError"]["sensorState"]["hasStartPermission"],
            result.data["sensorOrError"]["sensorState"]["hasStopPermission"],
            result.data["sensorOrError"]["hasCursorUpdatePermissions"],
        )

    def graphql_stop_schedule(self, context, schedule_id: str):
        result = execute_dagster_graphql(
            context,
            STOP_SCHEDULES_QUERY,
            variables={"id": schedule_id},
        )
        assert result.data
        return result.data["stopRunningSchedule"]["__typename"]

    def _create_started_schedule_state(self, context, schedule_name: str):
        selector = infer_schedule_selector(context, schedule_name)
        schedule = context.get_schedule(ScheduleSelector.from_graphql_input(selector))
        stored_state = context.instance.start_schedule(schedule)
        return CompoundID(
            remote_origin_id=stored_state.instigator_origin_id,
            selector_id=stored_state.selector_id,
        ).to_string()

    def graphql_start_sensor(self, context, sensor_name: str):
        selector = infer_sensor_selector(context, sensor_name)
        result = execute_dagster_graphql(
            context,
            START_SENSORS_QUERY,
            variables={
                "sensorSelector": selector,
            },
        )
        assert result.data
        return result.data["startSensor"]["__typename"]

    def graphql_update_sensor_cursor(self, context, sensor_name: str):
        selector = infer_sensor_selector(context, sensor_name)
        result = execute_dagster_graphql(
            context,
            """
            mutation SetSensorCursorMutation($sensorSelector: SensorSelector!, $cursor: String) {
                setSensorCursor(sensorSelector: $sensorSelector, cursor: $cursor) {
                    __typename
                    ... on Sensor {
                        name
                    }
                    ... on UnauthorizedError {
                        message
                    }
                }
            }
            """,
            variables={
                "sensorSelector": selector,
                "cursor": "new_cursor_value",
            },
        )
        assert result.data
        return result.data["setSensorCursor"]["__typename"]

    def graphql_delete_run(self, context, run_id: str):
        result = execute_dagster_graphql(context, DELETE_RUN_MUTATION, variables={"runId": run_id})
        assert result.data
        return result.data["deletePipelineRun"]["__typename"]

    def graphql_get_asset_wipe_permission(self, context, asset_key: AssetKey):
        result = execute_dagster_graphql(
            context,
            GET_ASSET_PERMISSIONS_QUERY,
            variables={"assetKey": asset_key.to_graphql_input()},
        )
        assert result.data
        if result.data["assetNodeOrError"]["__typename"] == "AssetNode":
            return result.data["assetNodeOrError"]["hasWipePermission"]
        return None

    def graphql_wipe_asset(self, context, asset_key: AssetKey):
        result = execute_dagster_graphql(
            context,
            WIPE_ASSETS,
            variables={"assetPartitionRanges": [{"assetKey": asset_key.to_graphql_input()}]},
        )
        assert result.data
        return result.data["wipeAssets"]["__typename"]

    def _create_run(self, context, job_name):
        remote_job = context.get_full_job(
            JobSelector.from_graphql_input(infer_job_selector(context, job_name))
        )
        repository = context.get_code_location(
            remote_job.repository_handle.location_name
        ).get_repository(remote_job.repository_handle.repository_name)
        return context.instance.create_run(
            job_name=job_name,
            run_id=None,
            run_config=None,
            status=None,
            tags=None,
            root_run_id=None,
            parent_run_id=None,
            step_keys_to_execute=None,
            execution_plan_snapshot=None,
            job_snapshot=None,
            parent_job_snapshot=None,
            asset_selection=None,
            asset_check_selection=None,
            resolved_op_selection=None,
            op_selection=None,
            job_code_origin=None,
            remote_job_origin=remote_job.get_remote_origin(),
            asset_graph=repository.asset_graph,
        )

    def _create_job_backfill(self, context: WorkspaceRequestContext, job_name: str):
        partition_set = self.get_partition_set_by_job_name(context, job_name)
        assert partition_set is not None
        backfill_id = make_new_backfill_id()
        partition_names = ["a", "b", "c"]
        backfill = PartitionBackfill(
            backfill_id=backfill_id,
            partition_set_origin=partition_set.get_remote_origin(),
            status=BulkActionStatus.REQUESTED,
            partition_names=partition_names,
            from_failure=False,
            reexecution_steps=None,
            tags={},
            backfill_timestamp=time.time(),
            asset_selection=None,
            title=None,
            description=None,
            run_config=None,
        )
        context.instance.add_backfill(backfill)
        return backfill_id

    def _create_asset_backfill(self, context: WorkspaceRequestContext, asset_keys: list[AssetKey]):
        backfill_id = make_new_backfill_id()
        backfill = PartitionBackfill.from_asset_partitions(
            backfill_id=backfill_id,
            asset_graph=context.asset_graph,
            backfill_timestamp=time.time(),
            tags={},
            asset_selection=asset_keys,
            partition_names=None,
            dynamic_partitions_store=context.instance,
            all_partitions=True,
            title=None,
            description=None,
            run_config=None,
        )
        context.instance.add_backfill(backfill)
        return backfill_id

    def _get_type_for_handle(self, def_handle: str | AssetKey) -> type[RemoteDefinition]:
        if isinstance(def_handle, AssetKey):
            return RemoteAssetNode
        if def_handle.endswith("_job"):
            return RemoteJob
        elif def_handle.endswith("_schedule"):
            return RemoteSchedule
        elif def_handle.endswith("_sensor"):
            return RemoteSensor

        raise Exception(f"Unknown definition handle {def_handle}")

    def _create_run_with_asset_check_selection(
        self, context, asset_selection: list[AssetKey], asset_check_selection: list[AssetCheckKey]
    ):
        return context.instance.create_run(
            job_name=IMPLICIT_ASSET_JOB_NAME,
            run_id=None,
            run_config=None,
            status=None,
            tags=None,
            root_run_id=None,
            parent_run_id=None,
            step_keys_to_execute=None,
            execution_plan_snapshot=None,
            job_snapshot=None,
            parent_job_snapshot=None,
            asset_selection=set(asset_selection),
            asset_check_selection=set(asset_check_selection),
            resolved_op_selection=None,
            op_selection=None,
            job_code_origin=None,
            remote_job_origin=None,
            asset_graph=None,
        )

    def test_definitions_are_owned(self, graphql_context: WorkspaceRequestContext):
        for def_handle in self.get_owned_definitions():
            def_type = self._get_type_for_handle(def_handle)
            selector = self._get_selector(graphql_context, def_type, def_handle)
            assert graphql_context.has_permission_for_selector(
                Permissions.LAUNCH_PIPELINE_EXECUTION, selector
            )
        for def_handle in self.get_unowned_definitions():
            def_type = self._get_type_for_handle(def_handle)
            selector = self._get_selector(graphql_context, def_type, def_handle)
            assert not graphql_context.has_permission_for_selector(
                Permissions.LAUNCH_PIPELINE_EXECUTION, selector
            )

    def test_get_owners_for_nonexistent_definitions(self, graphql_context: WorkspaceRequestContext):
        """get_owners_for_selector should return [] for selectors referencing
        definitions that don't exist in the workspace, rather than erroring.
        """
        # Get a valid location/repo name from the context
        code_location = graphql_context.code_locations[0]
        repository = next(iter(code_location.get_repositories().values()))
        location_name = code_location.name
        repository_name = repository.name

        # Non-existent asset key
        assert graphql_context.get_owners_for_selector(AssetKey(["does_not_exist"])) == []

        # Non-existent asset check key (asset doesn't exist)
        assert (
            graphql_context.get_owners_for_selector(
                AssetCheckKey(AssetKey(["does_not_exist"]), "some_check")
            )
            == []
        )

        # Non-existent job
        assert (
            graphql_context.get_owners_for_selector(
                JobSelector(
                    location_name=location_name,
                    repository_name=repository_name,
                    job_name="nonexistent_job",
                )
            )
            == []
        )

        # Implicit asset job name (should return [] even if it exists, since ownership
        # is determined by the assets, not the implicit job)
        assert (
            graphql_context.get_owners_for_selector(
                JobSelector(
                    location_name=location_name,
                    repository_name=repository_name,
                    job_name=IMPLICIT_ASSET_JOB_NAME,
                )
            )
            == []
        )

        # Non-existent schedule
        assert (
            graphql_context.get_owners_for_selector(
                ScheduleSelector(
                    location_name=location_name,
                    repository_name=repository_name,
                    schedule_name="nonexistent_schedule",
                )
            )
            == []
        )

        # Non-existent sensor
        assert (
            graphql_context.get_owners_for_selector(
                SensorSelector(
                    location_name=location_name,
                    repository_name=repository_name,
                    sensor_name="nonexistent_sensor",
                )
            )
            == []
        )

    def test_asset_backfill_launch_permissions(self, graphql_context: WorkspaceRequestContext):
        typename, _ = self.graphql_launch_asset_backfill(
            graphql_context, [AssetKey(["owned_partitioned_asset"])]
        )
        assert typename == "LaunchBackfillSuccess"

        typename, _ = self.graphql_launch_asset_backfill(
            graphql_context, [AssetKey(["unowned_partitioned_asset"])]
        )
        assert typename == "UnauthorizedError"

        typename, _ = self.graphql_launch_asset_backfill(
            graphql_context,
            [AssetKey(["owned_partitioned_asset"]), AssetKey(["unowned_partitioned_asset"])],
        )
        assert typename == "UnauthorizedError"

    def test_asset_backfill_cancel_permissions(self, graphql_context: WorkspaceRequestContext):
        backfill_id_a = self._create_asset_backfill(
            graphql_context, [AssetKey(["owned_partitioned_asset"])]
        )
        backfill_id_b = self._create_asset_backfill(
            graphql_context, [AssetKey(["unowned_partitioned_asset"])]
        )
        assert backfill_id_a
        assert backfill_id_b

        assert (
            self.graphql_cancel_backfill(graphql_context, backfill_id_a) == "CancelBackfillSuccess"
        )
        assert self.graphql_cancel_backfill(graphql_context, backfill_id_b) == "UnauthorizedError"

    def test_asset_run_launch_permissions(self, graphql_context: WorkspaceRequestContext):
        assert (
            self.graphql_launch_asset_run(graphql_context, [AssetKey(["owned_asset"])])
            == "LaunchRunSuccess"
        )
        assert (
            self.graphql_launch_asset_run(graphql_context, [AssetKey(["unowned_asset"])])
            == "UnauthorizedError"
        )
        assert (
            self.graphql_launch_asset_run(
                graphql_context, [AssetKey(["owned_asset"]), AssetKey(["unowned_asset"])]
            )
            == "UnauthorizedError"
        )

    def test_job_run_launch_permissions(self, graphql_context: WorkspaceRequestContext):
        can_launch, _ = self.graphql_get_job_permissions(graphql_context, "owned_job")
        assert can_launch
        typename, _ = self.graphql_launch_job_run(graphql_context, "owned_job")
        assert typename == "LaunchRunSuccess"

        can_launch, _ = self.graphql_get_job_permissions(graphql_context, "unowned_job")
        assert not can_launch
        typename, _ = self.graphql_launch_job_run(graphql_context, "unowned_job")
        assert typename == "UnauthorizedError"

    def test_implicit_asset_job_run_launch_permissions(
        self, graphql_context: WorkspaceRequestContext
    ):
        can_launch, _ = self.graphql_get_job_permissions(
            graphql_context, "__ASSET_JOB", asset_selection=[AssetKey(["owned_asset"])]
        )
        assert can_launch
        typename = self.graphql_launch_asset_run(
            graphql_context, asset_selection=[AssetKey(["owned_asset"])]
        )
        assert typename == "LaunchRunSuccess"

        can_launch, _ = self.graphql_get_job_permissions(
            graphql_context, "__ASSET_JOB", asset_selection=[AssetKey(["unowned_asset"])]
        )
        assert not can_launch
        typename = self.graphql_launch_asset_run(
            graphql_context, asset_selection=[AssetKey(["unowned_asset"])]
        )
        assert typename == "UnauthorizedError"

    def test_job_reexecution_permissions(self, graphql_context: WorkspaceRequestContext):
        _, can_reexecute = self.graphql_get_job_permissions(graphql_context, "owned_job")
        assert can_reexecute
        run_a = self._create_run(graphql_context, "owned_job")
        _, can_reexecute = self.graphql_get_job_permissions(graphql_context, "unowned_job")
        assert not can_reexecute
        run_b = self._create_run(graphql_context, "unowned_job")
        assert (
            self.graphql_launch_job_reexecution(graphql_context, "owned_job", run_a.run_id)
            == "LaunchRunSuccess"
        )
        assert (
            self.graphql_launch_job_reexecution(graphql_context, "unowned_job", run_b.run_id)
            == "UnauthorizedError"
        )

    def test_start_schedule_permissions(self, graphql_context: WorkspaceRequestContext):
        has_start_a, _ = self.graphql_get_schedule_permissions(graphql_context, "owned_schedule")
        assert has_start_a
        typename, _ = self.graphql_start_schedule(graphql_context, "owned_schedule")
        assert typename == "ScheduleStateResult"

        has_start_b, _ = self.graphql_get_schedule_permissions(graphql_context, "unowned_schedule")
        assert not has_start_b
        typename, _ = self.graphql_start_schedule(graphql_context, "unowned_schedule")
        assert typename == "UnauthorizedError"

    def test_stop_schedule_permissions(self, graphql_context: WorkspaceRequestContext):
        schedule_a_id = self._create_started_schedule_state(graphql_context, "owned_schedule")
        schedule_b_id = self._create_started_schedule_state(graphql_context, "unowned_schedule")
        _, has_stop_a = self.graphql_get_schedule_permissions(graphql_context, "owned_schedule")
        assert has_stop_a
        _, has_stop_b = self.graphql_get_schedule_permissions(graphql_context, "unowned_schedule")
        assert not has_stop_b
        assert self.graphql_stop_schedule(graphql_context, schedule_a_id) == "ScheduleStateResult"
        assert self.graphql_stop_schedule(graphql_context, schedule_b_id) == "UnauthorizedError"

    def test_sensor_permissions(self, graphql_context: WorkspaceRequestContext):
        has_start_a, has_stop_a, has_cursor_update_a = self.graphql_get_sensor_permissions(
            graphql_context, "owned_sensor"
        )
        assert has_start_a
        assert has_stop_a
        assert has_cursor_update_a
        assert self.graphql_start_sensor(graphql_context, "owned_sensor") == "Sensor"
        has_start_b, has_stop_b, has_cursor_update_b = self.graphql_get_sensor_permissions(
            graphql_context, "unowned_sensor"
        )
        assert not has_start_b
        assert not has_stop_b
        assert not has_cursor_update_b
        assert self.graphql_start_sensor(graphql_context, "unowned_sensor") == "UnauthorizedError"
        assert self.graphql_update_sensor_cursor(graphql_context, "owned_sensor") == "Sensor"
        assert (
            self.graphql_update_sensor_cursor(graphql_context, "unowned_sensor")
            == "UnauthorizedError"
        )

    def test_run_deletion_permissions(self, graphql_context: WorkspaceRequestContext):
        run_a = self._create_run(graphql_context, "owned_job")
        run_b = self._create_run(graphql_context, "unowned_job")
        assert self.graphql_delete_run(graphql_context, run_a.run_id) == "DeletePipelineRunSuccess"
        assert self.graphql_delete_run(graphql_context, run_b.run_id) == "UnauthorizedError"

    def test_job_backfill_launch_permissions(self, graphql_context: WorkspaceRequestContext):
        can_launch, _ = self.graphql_has_job_backfill_permissions(
            graphql_context, "owned_partitioned_job"
        )
        assert can_launch
        typename, _ = self.graphql_launch_job_backfill(graphql_context, "owned_partitioned_job")
        assert typename == "LaunchBackfillSuccess"

        can_launch, _ = self.graphql_has_job_backfill_permissions(
            graphql_context, "unowned_partitioned_job"
        )
        assert not can_launch
        typename, _ = self.graphql_launch_job_backfill(graphql_context, "unowned_partitioned_job")
        assert typename == "UnauthorizedError"

    def test_job_backfill_cancel_permissions(self, graphql_context: WorkspaceRequestContext):
        backfill_id_a = self._create_job_backfill(graphql_context, "owned_partitioned_job")
        backfill_id_b = self._create_job_backfill(graphql_context, "unowned_partitioned_job")
        _, can_cancel = self.graphql_has_job_backfill_permissions(
            graphql_context, "owned_partitioned_job"
        )
        assert can_cancel
        _, can_cancel = self.graphql_has_job_backfill_permissions(
            graphql_context, "unowned_partitioned_job"
        )
        assert not can_cancel
        assert (
            self.graphql_cancel_backfill(graphql_context, backfill_id_a) == "CancelBackfillSuccess"
        )
        assert self.graphql_cancel_backfill(graphql_context, backfill_id_b) == "UnauthorizedError"

    def test_asset_check_run_launch_permissions(self, graphql_context: WorkspaceRequestContext):
        assert (
            self.graphql_launch_asset_job(
                graphql_context, [], [AssetCheckKey(AssetKey(["owned_asset"]), "owned_asset_check")]
            )
            == "LaunchRunSuccess"
        )
        assert (
            self.graphql_launch_asset_job(
                graphql_context,
                [],
                [AssetCheckKey(AssetKey(["unowned_asset"]), "unowned_asset_check")],
            )
            == "UnauthorizedError"
        )

    def test_mixed_asset_and_check_run_launch_permissions(
        self, graphql_context: WorkspaceRequestContext
    ):
        assert (
            self.graphql_launch_asset_job(
                graphql_context,
                [AssetKey(["owned_asset"])],
                [AssetCheckKey(AssetKey(["owned_asset"]), "owned_asset_check")],
            )
            == "LaunchRunSuccess"
        )
        assert (
            self.graphql_launch_asset_job(
                graphql_context,
                [AssetKey(["unowned_asset"])],
                [AssetCheckKey(AssetKey(["unowned_asset"]), "unowned_asset_check")],
            )
            == "UnauthorizedError"
        )
        assert (
            self.graphql_launch_asset_job(
                graphql_context,
                [AssetKey(["owned_asset"])],
                [AssetCheckKey(AssetKey(["unowned_asset"]), "unowned_asset_check")],
            )
            == "UnauthorizedError"
        )
        assert (
            self.graphql_launch_asset_job(
                graphql_context,
                [AssetKey(["unowned_asset"])],
                [AssetCheckKey(AssetKey(["owned_asset"]), "owned_asset_check")],
            )
            == "UnauthorizedError"
        )

    def test_asset_check_reexecution_permissions(self, graphql_context: WorkspaceRequestContext):
        run_with_owned = self._create_run_with_asset_check_selection(
            graphql_context,
            asset_selection=[AssetKey(["owned_asset"])],
            asset_check_selection=[AssetCheckKey(AssetKey(["owned_asset"]), "owned_asset_check")],
        )

        run_with_unowned = self._create_run_with_asset_check_selection(
            graphql_context,
            asset_selection=[AssetKey(["unowned_asset"])],
            asset_check_selection=[
                AssetCheckKey(AssetKey(["unowned_asset"]), "unowned_asset_check")
            ],
        )
        assert (
            self.graphql_launch_job_reexecution(
                graphql_context,
                IMPLICIT_ASSET_JOB_NAME,
                run_with_owned.run_id,
                asset_selection=[AssetKey(["owned_asset"])],
                asset_check_selection=[
                    AssetCheckKey(AssetKey(["owned_asset"]), "owned_asset_check")
                ],
            )
            == "LaunchRunSuccess"
        )
        assert (
            self.graphql_launch_job_reexecution(
                graphql_context,
                IMPLICIT_ASSET_JOB_NAME,
                run_with_unowned.run_id,
                asset_selection=[AssetKey(["unowned_asset"])],
                asset_check_selection=[
                    AssetCheckKey(AssetKey(["unowned_asset"]), "unowned_asset_check")
                ],
            )
            == "UnauthorizedError"
        )

    def test_asset_wipe_permissions(self, graphql_context: WorkspaceRequestContext):
        instance = graphql_context.instance
        materialize([owned_asset], instance=instance)
        materialize([unowned_asset], instance=instance)

        assert (
            len(instance.fetch_materializations(AssetKey(["owned_asset"]), limit=10).records) == 1
        )
        assert (
            len(instance.fetch_materializations(AssetKey(["unowned_asset"]), limit=10).records) == 1
        )

        assert self.graphql_get_asset_wipe_permission(graphql_context, AssetKey(["owned_asset"]))
        assert not self.graphql_get_asset_wipe_permission(
            graphql_context, AssetKey(["unowned_asset"])
        )

        assert (
            self.graphql_wipe_asset(graphql_context, AssetKey(["owned_asset"]))
            == "AssetWipeSuccess"
        )
        assert (
            self.graphql_wipe_asset(graphql_context, AssetKey(["unowned_asset"]))
            == "UnauthorizedError"
        )


class TestDefinitionOwnerPermissions(
    BaseDefinitionOwnerPermissionsTestSuite,
    BaseTestSuite,
):
    @contextmanager
    def yield_graphql_context(
        self, class_scoped_context
    ) -> Generator[WorkspaceRequestContext, None, None]:
        with super().yield_graphql_context(class_scoped_context) as context:
            with self.setup_permissions(context):
                yield context

    @contextmanager
    def setup_permissions(self, context: WorkspaceRequestContext) -> Iterator[None]:
        _did_check = {}

        def _mock_unpermitted(permission: str, *args) -> bool:
            _did_check[permission] = True
            return False

        def _mock_owner_permissions(
            permission: str,
            owners: Sequence[str],
        ) -> bool:
            return (
                len(
                    set(owners).intersection(
                        set(
                            [
                                "test@elementl.com",
                                "team:foo",
                            ]
                        )
                    )
                )
                > 0
            )

        def _mock_selector_ownership(
            permission: str,
            selector: JobSelector | ScheduleSelector | SensorSelector | AssetKey | AssetCheckKey,
        ) -> bool:
            owned_defs = self.get_owned_definitions()
            _did_check[permission] = True
            if isinstance(selector, AssetKey):
                return selector in owned_defs
            elif isinstance(selector, AssetCheckKey):
                # Asset checks inherit permissions from their underlying asset
                return selector.asset_key in owned_defs
            elif isinstance(selector, JobSelector):
                return selector.job_name in owned_defs
            elif isinstance(selector, ScheduleSelector):
                return selector.schedule_name in owned_defs
            elif isinstance(selector, SensorSelector):
                return selector.sensor_name in owned_defs
            return False

        def _mock_permission_check(permission):
            return _did_check.get(permission) or False

        with (
            mock.patch.object(
                context,
                "has_permission",
                side_effect=_mock_unpermitted,
            ),
            mock.patch.object(
                context,
                "has_permission_for_location",
                side_effect=_mock_unpermitted,
            ),
            mock.patch.object(
                context,
                "viewer_has_any_owner_definition_permissions",
                side_effect=lambda: True,
            ),
            mock.patch.object(
                context, "has_permission_for_owners", side_effect=_mock_owner_permissions
            ),
            mock.patch.object(
                context, "has_permission_for_selector", side_effect=_mock_selector_ownership
            ),
            mock.patch.object(
                context, "was_permission_checked", side_effect=_mock_permission_check
            ),
        ):
            yield



from typing import TYPE_CHECKING, Union

import dagster._check as check
from dagster._core.instance import DagsterInstance
from dagster._core.storage.dagster_run import DagsterRun

from dagster_graphql.implementation.external import get_full_remote_job_or_raise
from dagster_graphql.implementation.utils import JobSubsetSelector, UserFacingGraphQLError
from dagster_graphql.schema.util import ResolveInfo

if TYPE_CHECKING:
    from dagster_graphql.schema.pipelines.pipeline_ref import GrapheneUnknownPipeline
    from dagster_graphql.schema.pipelines.snapshot import GraphenePipelineSnapshot


async def get_job_snapshot_or_error_from_job_selector(
    graphene_info: ResolveInfo, job_selector: JobSubsetSelector
) -> "GraphenePipelineSnapshot":
    from dagster_graphql.schema.pipelines.snapshot import GraphenePipelineSnapshot

    check.inst_param(job_selector, "pipeline_selector", JobSubsetSelector)
    return GraphenePipelineSnapshot(await get_full_remote_job_or_raise(graphene_info, job_selector))


def get_job_snapshot_or_error_from_snapshot_id(
    graphene_info: ResolveInfo, snapshot_id: str
) -> "GraphenePipelineSnapshot":
    check.str_param(snapshot_id, "snapshot_id")

    return _get_job_snapshot_from_instance(graphene_info.context.instance, snapshot_id)


async def get_job_snapshot_or_error_from_snap_or_selector(
    graphene_info: ResolveInfo,
    job_selector: JobSubsetSelector,
    snapshot_id: str,
):
    from dagster_graphql.schema.pipelines.snapshot import GraphenePipelineSnapshot

    if graphene_info.context.instance.has_job_snapshot(snapshot_id):
        job_snapshot = graphene_info.context.instance.get_historical_job(snapshot_id)
        if job_snapshot:
            return GraphenePipelineSnapshot(job_snapshot)

    return await get_job_snapshot_or_error_from_job_selector(graphene_info, job_selector)


# extracted this out to test
def _get_job_snapshot_from_instance(
    instance: DagsterInstance, snapshot_id: str
) -> "GraphenePipelineSnapshot":
    from dagster_graphql.schema.errors import GraphenePipelineSnapshotNotFoundError
    from dagster_graphql.schema.pipelines.snapshot import GraphenePipelineSnapshot

    if not instance.has_job_snapshot(snapshot_id):
        raise UserFacingGraphQLError(GraphenePipelineSnapshotNotFoundError(snapshot_id))

    historical_pipeline = instance.get_historical_job(snapshot_id)

    if not historical_pipeline:
        # Either a temporary error or it has been deleted in the interim
        raise UserFacingGraphQLError(GraphenePipelineSnapshotNotFoundError(snapshot_id))

    return GraphenePipelineSnapshot(historical_pipeline)


def get_job_reference_or_raise(
    graphene_info: ResolveInfo, dagster_run: DagsterRun
) -> Union["GraphenePipelineSnapshot", "GrapheneUnknownPipeline"]:
    """Returns a PipelineReference or raises a UserFacingGraphQLError if a pipeline
    reference cannot be retrieved based on the run, e.g, a UserFacingGraphQLError that wraps an
    InvalidSubsetError.
    """
    from dagster_graphql.schema.pipelines.pipeline_ref import GrapheneUnknownPipeline

    check.inst_param(dagster_run, "pipeline_run", DagsterRun)
    op_selection = (
        list(dagster_run.resolved_op_selection) if dagster_run.resolved_op_selection else None
    )

    if dagster_run.job_snapshot_id is None:
        return GrapheneUnknownPipeline(dagster_run.job_name, op_selection)

    return _get_job_snapshot_from_instance(
        graphene_info.context.instance, dagster_run.job_snapshot_id
    )



import dagster._check as check
import graphene
from dagster._core.remote_representation.represented import RepresentedJob

from dagster_graphql.schema.errors import (
    GraphenePipelineNotFoundError,
    GraphenePipelineSnapshotNotFoundError,
    GraphenePythonError,
)
from dagster_graphql.schema.pipelines.pipeline import (
    GrapheneIPipelineSnapshot,
    GrapheneIPipelineSnapshotMixin,
)
from dagster_graphql.schema.pipelines.pipeline_ref import GraphenePipelineReference
from dagster_graphql.schema.solids import GrapheneSolidContainer


class GraphenePipelineSnapshot(GrapheneIPipelineSnapshotMixin, graphene.ObjectType):
    class Meta:  # pyright: ignore[reportIncompatibleVariableOverride]
        interfaces = (GrapheneSolidContainer, GrapheneIPipelineSnapshot, GraphenePipelineReference)
        name = "PipelineSnapshot"

    def __init__(self, represented_job: RepresentedJob):
        super().__init__()
        self._represented_job = check.inst_param(
            represented_job, "represented_pipeline", RepresentedJob
        )

    def get_represented_job(self) -> RepresentedJob:
        return self._represented_job


class GraphenePipelineSnapshotOrError(graphene.Union):
    class Meta:
        types = (
            GraphenePipelineNotFoundError,
            GraphenePipelineSnapshot,
            GraphenePipelineSnapshotNotFoundError,
            GraphenePythonError,
        )
        name = "PipelineSnapshotOrError"



import os
import time
from typing import Any
from uuid import uuid4

import pytest
from dagster._core.definitions.reconstruct import ReconstructableRepository
from dagster._core.execution.api import execute_job
from dagster._core.storage.dagster_run import DagsterRunStatus
from dagster._core.test_utils import create_run_for_test
from dagster._core.utils import make_new_run_id
from dagster._core.workspace.context import WorkspaceRequestContext
from dagster._grpc.types import CancelExecutionRequest
from dagster._utils import file_relative_path, safe_tempfile_path
from dagster_graphql.client.query import LAUNCH_PIPELINE_EXECUTION_MUTATION
from dagster_graphql.test.utils import execute_dagster_graphql, infer_job_selector

from dagster_graphql_tests.graphql.graphql_context_test_suite import (
    GraphQLContextVariant,
    ReadonlyGraphQLContextTestMatrix,
    make_graphql_context_test_suite,
)

RUN_CANCELLATION_QUERY = """
mutation($runId: String!, $terminatePolicy: TerminateRunPolicy) {
  terminatePipelineExecution(runId: $runId, terminatePolicy:$terminatePolicy){
    __typename
    ... on TerminateRunSuccess{
      run {
        runId
      }
    }
    ... on TerminateRunFailure {
      run {
        runId
      }
      message
    }
    ... on RunNotFoundError {
      runId
    }
    ... on PythonError {
      message
      stack
    }
  }
}
"""

RUNS_CANCELLATION_QUERY = """
mutation TerminateRuns($runIds: [String!]!) {
  terminateRuns(runIds: $runIds) {
    __typename
    ... on PythonError{
      message
    }
    ... on TerminateRunsResult {
      terminateRunResults {
        __typename
        ... on TerminateRunSuccess {
            run {
                runId
            }
        }
        ... on RunNotFoundError {
            runId
        }
        ... on TerminateRunFailure {
            run {
                runId
            }
            message
        }
      }
    }
  }
}
"""

BULK_TERMINATION_PERMISSIONS_QUERY = """
query canBulkTerminate {
    canBulkTerminate
}
"""

# Legacy query (once published on docs site), that we should try to continue supporting
BACKCOMPAT_LEGACY_TERMINATE_PIPELINE = """
mutation TerminatePipeline($runId: String!) {
  terminatePipelineExecution(runId: $runId){
    __typename
    ... on TerminatePipelineExecutionSuccess{
      run {
        runId
      }
    }
    ... on TerminatePipelineExecutionFailure {
      message
    }
    ... on PipelineRunNotFoundError {
      runId
    }
    ... on PythonError {
      message
      stack
    }
  }
}
"""


QueuedRunCoordinatorTestSuite: Any = make_graphql_context_test_suite(
    context_variants=[GraphQLContextVariant.sqlite_with_queued_run_coordinator_managed_grpc_env()]
)


class TestQueuedRunTermination(QueuedRunCoordinatorTestSuite):
    def test_cancel_queued_run(self, graphql_context: WorkspaceRequestContext):
        selector = infer_job_selector(graphql_context, "infinite_loop_job")
        with safe_tempfile_path() as path:
            result = execute_dagster_graphql(
                graphql_context,
                LAUNCH_PIPELINE_EXECUTION_MUTATION,
                variables={
                    "executionParams": {
                        "selector": selector,
                        "mode": "default",
                        "runConfigData": {"ops": {"loop": {"config": {"file": path}}}},
                    }
                },
            )

            assert not result.errors
            assert result.data

            # just test existence
            assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"
            run_id = result.data["launchPipelineExecution"]["run"]["runId"]

            run = graphql_context.instance.get_run_by_id(run_id)
            assert run and run.status == DagsterRunStatus.QUEUED

            result = execute_dagster_graphql(
                graphql_context, RUN_CANCELLATION_QUERY, variables={"runId": run_id}
            )
            assert (
                result.data["terminatePipelineExecution"]["__typename"] == "TerminateRunSuccess"
            ), str(result.data)

    def test_cancel_runs(self, graphql_context: WorkspaceRequestContext):
        selector = infer_job_selector(graphql_context, "infinite_loop_job")
        with safe_tempfile_path() as path:
            result = execute_dagster_graphql(
                graphql_context,
                LAUNCH_PIPELINE_EXECUTION_MUTATION,
                variables={
                    "executionParams": {
                        "selector": selector,
                        "mode": "default",
                        "runConfigData": {"ops": {"loop": {"config": {"file": path}}}},
                    }
                },
            )

            assert not result.errors
            assert result.data

            # just test existence
            assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"
            run_id = result.data["launchPipelineExecution"]["run"]["runId"]

            run = graphql_context.instance.get_run_by_id(run_id)
            assert run and run.status == DagsterRunStatus.QUEUED

            result = execute_dagster_graphql(
                graphql_context,
                RUNS_CANCELLATION_QUERY,
                variables={"runIds": [run_id, "nonexistent_id"]},
            )
            assert result.data["terminateRuns"]["__typename"] == "TerminateRunsResult"
            assert (
                result.data["terminateRuns"]["terminateRunResults"][0]["__typename"]
                == "TerminateRunSuccess"
            )
            assert (
                result.data["terminateRuns"]["terminateRunResults"][1]["__typename"]
                == "RunNotFoundError"
            )
            assert (
                result.data["terminateRuns"]["terminateRunResults"][1]["runId"] == "nonexistent_id"
            )

    def test_force_cancel_queued_run(self, graphql_context: WorkspaceRequestContext):
        selector = infer_job_selector(graphql_context, "infinite_loop_job")
        with safe_tempfile_path() as path:
            result = execute_dagster_graphql(
                graphql_context,
                LAUNCH_PIPELINE_EXECUTION_MUTATION,
                variables={
                    "executionParams": {
                        "selector": selector,
                        "mode": "default",
                        "runConfigData": {"ops": {"loop": {"config": {"file": path}}}},
                    }
                },
            )

            assert not result.errors
            assert result.data

            # just test existence
            assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"
            run_id = result.data["launchPipelineExecution"]["run"]["runId"]

            run = graphql_context.instance.get_run_by_id(run_id)
            assert run and run.status == DagsterRunStatus.QUEUED

            result = execute_dagster_graphql(
                graphql_context,
                RUN_CANCELLATION_QUERY,
                variables={
                    "runId": run_id,
                    "terminatePolicy": "MARK_AS_CANCELED_IMMEDIATELY",
                },
            )
            assert result.data["terminatePipelineExecution"]["__typename"] == "TerminateRunSuccess"


RunTerminationTestSuite: Any = make_graphql_context_test_suite(
    context_variants=[GraphQLContextVariant.sqlite_with_default_run_launcher_managed_grpc_env()]
)


def _exception_terminate(_run_id):
    raise Exception("FAILED TO TERMINATE")


def _return_fail_terminate(_run_id):
    return False


class TestTerminationReadonly(ReadonlyGraphQLContextTestMatrix):
    def test_termination_permission_failure(self, graphql_context: WorkspaceRequestContext):
        run_id = create_run_for_test(graphql_context.instance).run_id
        result = execute_dagster_graphql(
            graphql_context, RUN_CANCELLATION_QUERY, variables={"runId": run_id}
        )
        assert not result.errors
        assert result.data

        # just test existence
        assert result.data["terminatePipelineExecution"]["__typename"] == "TerminateRunFailure"
        assert "do not have permission" in result.data["terminatePipelineExecution"]["message"]

    def test_cancel_runs_permission_failure(self, graphql_context: WorkspaceRequestContext):
        run_id = create_run_for_test(graphql_context.instance).run_id
        result = execute_dagster_graphql(
            graphql_context, RUNS_CANCELLATION_QUERY, variables={"runIds": [run_id]}
        )
        assert not result.errors
        assert result.data

        # just test existence
        assert result.data["terminateRuns"]["__typename"] == "TerminateRunsResult"
        assert len(result.data["terminateRuns"]["terminateRunResults"]) == 1
        assert (
            result.data["terminateRuns"]["terminateRunResults"][0]["__typename"]
            == "TerminateRunFailure"
        )
        assert (
            "do not have permission"
            in result.data["terminateRuns"]["terminateRunResults"][0]["message"]
        )

    def test_cancel_runs_permission_failure_non_existent_run(
        self, graphql_context: WorkspaceRequestContext
    ):
        run_id = make_new_run_id()
        result = execute_dagster_graphql(
            graphql_context, RUNS_CANCELLATION_QUERY, variables={"runIds": [run_id]}
        )

        assert (
            result.data["terminateRuns"]["terminateRunResults"][0]["__typename"]
            == "UnauthorizedError"
        )

    def test_no_bulk_terminate_permission(self, graphql_context: WorkspaceRequestContext):
        result = execute_dagster_graphql(graphql_context, BULK_TERMINATION_PERMISSIONS_QUERY)
        assert not result.errors
        assert result.data

        assert result.data["canBulkTerminate"] is False


class TestRunVariantTermination(RunTerminationTestSuite):
    def test_basic_termination(self, graphql_context: WorkspaceRequestContext):
        selector = infer_job_selector(graphql_context, "infinite_loop_job")
        with safe_tempfile_path() as path:
            result = execute_dagster_graphql(
                graphql_context,
                LAUNCH_PIPELINE_EXECUTION_MUTATION,
                variables={
                    "executionParams": {
                        "selector": selector,
                        "mode": "default",
                        "runConfigData": {"ops": {"loop": {"config": {"file": path}}}},
                    }
                },
            )

            assert not result.errors
            assert result.data

            # just test existence
            assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"
            run_id = result.data["launchPipelineExecution"]["run"]["runId"]

            assert run_id

            # ensure the execution has happened
            while not os.path.exists(path):
                time.sleep(0.1)

            result = execute_dagster_graphql(
                graphql_context, RUN_CANCELLATION_QUERY, variables={"runId": run_id}
            )
            assert result.data["terminatePipelineExecution"]["__typename"] == "TerminateRunSuccess"

    def test_force_termination(self, graphql_context: WorkspaceRequestContext):
        selector = infer_job_selector(graphql_context, "infinite_loop_job")
        with safe_tempfile_path() as path:
            result = execute_dagster_graphql(
                graphql_context,
                LAUNCH_PIPELINE_EXECUTION_MUTATION,
                variables={
                    "executionParams": {
                        "selector": selector,
                        "mode": "default",
                        "runConfigData": {"ops": {"loop": {"config": {"file": path}}}},
                    }
                },
            )

            assert not result.errors
            assert result.data

            # just test existence
            assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"
            run_id = result.data["launchPipelineExecution"]["run"]["runId"]

            assert run_id

            # ensure the execution has happened
            while not os.path.exists(path):
                time.sleep(0.1)

            result = execute_dagster_graphql(
                graphql_context,
                RUN_CANCELLATION_QUERY,
                variables={
                    "runId": run_id,
                    "terminatePolicy": "MARK_AS_CANCELED_IMMEDIATELY",
                },
            )
            assert result.data["terminatePipelineExecution"]["__typename"] == "TerminateRunSuccess"

            instance = graphql_context.instance
            run = instance.get_run_by_id(run_id)
            assert run and run.status == DagsterRunStatus.CANCELED

    def test_run_not_found(self, graphql_context: WorkspaceRequestContext):
        random_run_id = str(uuid4())
        result = execute_dagster_graphql(
            graphql_context, RUN_CANCELLATION_QUERY, variables={"runId": random_run_id}
        )
        assert result.data["terminatePipelineExecution"]["__typename"] == "RunNotFoundError"

    @pytest.mark.parametrize(
        argnames=["new_terminate_method", "terminate_result"],
        argvalues=[
            [
                _return_fail_terminate,
                "TerminateRunFailure",
            ],
            [_exception_terminate, "PythonError"],
        ],
    )
    def test_terminate_failed(
        self, graphql_context: WorkspaceRequestContext, new_terminate_method, terminate_result
    ):
        selector = infer_job_selector(graphql_context, "infinite_loop_job")
        with safe_tempfile_path() as path:
            old_terminate = graphql_context.instance.run_launcher.terminate
            graphql_context.instance.run_launcher.terminate = new_terminate_method
            result = execute_dagster_graphql(
                graphql_context,
                LAUNCH_PIPELINE_EXECUTION_MUTATION,
                variables={
                    "executionParams": {
                        "selector": selector,
                        "mode": "default",
                        "runConfigData": {"ops": {"loop": {"config": {"file": path}}}},
                    }
                },
            )

            assert not result.errors
            assert result.data

            # just test existence
            assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"
            run_id = result.data["launchPipelineExecution"]["run"]["runId"]
            # ensure the execution has happened
            while not os.path.exists(path):
                time.sleep(0.1)

            result = execute_dagster_graphql(
                graphql_context, RUN_CANCELLATION_QUERY, variables={"runId": run_id}
            )
            assert result.data["terminatePipelineExecution"]["__typename"] == terminate_result, str(
                result.data
            )

            result = execute_dagster_graphql(
                graphql_context,
                RUN_CANCELLATION_QUERY,
                variables={
                    "runId": run_id,
                    "terminatePolicy": "MARK_AS_CANCELED_IMMEDIATELY",
                },
            )

            assert result.data["terminatePipelineExecution"]["__typename"] == "TerminateRunSuccess"

            assert result.data["terminatePipelineExecution"]["run"]["runId"] == run_id

            graphql_context.instance.run_launcher.terminate = old_terminate

            # Clean up the run process on the gRPC server
            code_location = graphql_context.code_locations[0]
            code_location.client.cancel_execution(CancelExecutionRequest(run_id=run_id))  # type: ignore

            run = graphql_context.instance.get_run_by_id(run_id)
            assert run and run.status == DagsterRunStatus.CANCELED

    def test_run_finished(self, graphql_context: WorkspaceRequestContext):
        instance = graphql_context.instance

        recon_job = ReconstructableRepository.for_file(
            file_relative_path(__file__, "repo.py"),
            "test_repo",
        ).get_reconstructable_job("noop_job")

        with execute_job(recon_job, instance=instance) as exec_result:
            assert exec_result.success
            assert exec_result.run_id

            time.sleep(0.05)  # guarantee execution finish

            result = execute_dagster_graphql(
                graphql_context,
                RUN_CANCELLATION_QUERY,
                variables={"runId": exec_result.run_id},
            )

            assert result.data["terminatePipelineExecution"]["__typename"] == "TerminateRunFailure"
            assert (
                "could not be terminated due to having status SUCCESS."
                in result.data["terminatePipelineExecution"]["message"]
            )

            # Still fails even if you change the terminate policy to fail immediately
            result = execute_dagster_graphql(
                graphql_context,
                RUN_CANCELLATION_QUERY,
                variables={
                    "runId": exec_result.run_id,
                    "terminatePolicy": "MARK_AS_CANCELED_IMMEDIATELY",
                },
            )

            assert result.data["terminatePipelineExecution"]["__typename"] == "TerminateRunFailure"
            assert (
                "could not be terminated due to having status SUCCESS."
                in result.data["terminatePipelineExecution"]["message"]
            )

    def test_backcompat_termination(self, graphql_context: WorkspaceRequestContext):
        selector = infer_job_selector(graphql_context, "infinite_loop_job")
        with safe_tempfile_path() as path:
            result = execute_dagster_graphql(
                graphql_context,
                LAUNCH_PIPELINE_EXECUTION_MUTATION,
                variables={
                    "executionParams": {
                        "selector": selector,
                        "mode": "default",
                        "runConfigData": {"ops": {"loop": {"config": {"file": path}}}},
                    }
                },
            )

            assert not result.errors
            assert result.data

            # just test existence
            assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"
            run_id = result.data["launchPipelineExecution"]["run"]["runId"]

            assert run_id

            # ensure the execution has happened
            while not os.path.exists(path):
                time.sleep(0.1)

            result = execute_dagster_graphql(
                graphql_context,
                BACKCOMPAT_LEGACY_TERMINATE_PIPELINE,
                variables={"runId": run_id},
            )
            assert result.data["terminatePipelineExecution"]["run"]["runId"] == run_id

    def test_has_bulk_terminate_permission(self, graphql_context: WorkspaceRequestContext):
        result = execute_dagster_graphql(graphql_context, BULK_TERMINATION_PERMISSIONS_QUERY)
        assert not result.errors
        assert result.data

        assert result.data["canBulkTerminate"] is True



import asyncio

from dagster._core.definitions.asset_health.asset_materialization_health import (
    AssetMaterializationHealthState,
)
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.partitions.subset import PartitionsSubset
from dagster._core.instance.types import CachingDynamicPartitionsLoader
from dagster._core.remote_representation.external_data import AssetNodeSnap
from dagster._core.storage.partition_status_cache import get_partition_subsets
from dagster._core.workspace.context import BaseWorkspaceRequestContext


async def _gen_asset_health_for_partition_subsets(
    context: BaseWorkspaceRequestContext,
    asset_key: AssetKey,
) -> AssetMaterializationHealthState | None:
    if not context.read_partition_subsets_from_asset_health():
        return None

    return await AssetMaterializationHealthState.gen(context, asset_key)


async def regenerate_and_check_partition_subsets(
    context: BaseWorkspaceRequestContext,
    asset_node_snap: AssetNodeSnap,
    dynamic_partitions_loader: CachingDynamicPartitionsLoader,
) -> tuple[
    PartitionsSubset[str] | None,
    PartitionsSubset[str] | None,
    PartitionsSubset[str] | None,
]:
    (
        asset_health_state,
        (materialized_partition_subset, failed_partition_subset, in_progress_subset),
    ) = await asyncio.gather(
        _gen_asset_health_for_partition_subsets(context, asset_node_snap.asset_key),
        get_partition_subsets(
            context.instance,
            context,
            asset_node_snap.asset_key,
            dynamic_partitions_loader,
            (
                asset_node_snap.partitions.get_partitions_definition()
                if asset_node_snap.partitions
                else None
            ),
        ),
    )

    if asset_health_state:
        # prefer sourcing materialized and failed subsets from the health object if possible
        materialized_partition_subset = (
            asset_health_state.materialized_subset.subset_value
            if asset_health_state.materialized_subset.is_partitioned
            else None
        )
        failed_partition_subset = (
            asset_health_state.failed_subset.subset_value
            if asset_health_state.failed_subset.is_partitioned
            else None
        )

    return (
        materialized_partition_subset,
        failed_partition_subset,
        in_progress_subset,
    )



import pytest
from dagster._core.utils import make_new_run_id
from dagster_graphql import DagsterGraphQLClientError

from dagster_graphql_tests.client_tests.conftest import MockClient, python_client_test_suite

RUN_ID = make_new_run_id()


@python_client_test_suite
def test_terminate_run_status_success(mock_client: MockClient):
    expected_result = None
    response = {"terminateRun": {"__typename": "TerminateRunSuccess", "run": expected_result}}
    mock_client.mock_gql_client.execute.return_value = response

    actual_result = mock_client.python_client.terminate_run(RUN_ID)
    assert actual_result == expected_result


@python_client_test_suite
def test_terminate_run_not_failure(mock_client: MockClient):
    error_type, error_message = "TerminateRunFailure", "Unable to terminate run"
    response = {"terminateRun": {"__typename": "TerminateRunFailure", "message": error_message}}
    mock_client.mock_gql_client.execute.return_value = response

    with pytest.raises(DagsterGraphQLClientError) as e:
        mock_client.python_client.terminate_run(RUN_ID)
    assert e.value.args == (error_type, error_message)


@python_client_test_suite
def test_terminate_run_not_found(mock_client: MockClient):
    error_type, error_message = "RunNotFoundError", f"Run Id {RUN_ID} not found"
    response = {"terminateRun": {"__typename": "RunNotFoundError", "runId": error_message}}

    with pytest.raises(DagsterGraphQLClientError) as e:
        mock_client.mock_gql_client.execute.return_value = response
        mock_client.python_client.terminate_run(RUN_ID)
    assert e.value.args == (error_type, error_message)


@python_client_test_suite
def test_terminate_run_python_error(mock_client: MockClient):
    error_type, error_message = "PythonError", "Unable to terminate run"
    response = {"terminateRun": {"__typename": "PythonError", "message": error_message}}

    with pytest.raises(DagsterGraphQLClientError) as e:
        mock_client.mock_gql_client.execute.return_value = response
        mock_client.python_client.terminate_run(RUN_ID)
    assert e.value.args == (error_type, error_message)



import time
from typing import Any

from dagster._core.errors import DagsterUserCodeUnreachableError
from dagster_graphql import ShutdownRepositoryLocationStatus
from dagster_graphql.client.client_queries import SHUTDOWN_REPOSITORY_LOCATION_MUTATION
from dagster_graphql.test.utils import execute_dagster_graphql, main_repo_location_name

from dagster_graphql_tests.graphql.graphql_context_test_suite import (
    GraphQLContextVariant,
    ReadonlyGraphQLContextTestMatrix,
    make_graphql_context_test_suite,
)

BaseTestSuite: Any = make_graphql_context_test_suite(
    context_variants=[GraphQLContextVariant.non_launchable_sqlite_instance_deployed_grpc_env()]
)


class TestShutdownRepositoryLocationReadOnly(ReadonlyGraphQLContextTestMatrix):
    def test_shutdown_repository_location_permission_failure(self, graphql_context):
        result = execute_dagster_graphql(
            graphql_context,
            SHUTDOWN_REPOSITORY_LOCATION_MUTATION,
            {"repositoryLocationName": main_repo_location_name()},
        )

        assert result
        assert result.data
        assert result.data["shutdownRepositoryLocation"]
        assert result.data["shutdownRepositoryLocation"]["__typename"] == "UnauthorizedError"


class TestShutdownRepositoryLocation(BaseTestSuite):
    def test_shutdown_repository_location(self, graphql_client, graphql_context):
        origin = next(iter(graphql_context.get_code_location_entries().values())).origin
        origin.create_client().heartbeat()

        result = graphql_client.shutdown_repository_location(main_repo_location_name())

        assert result.status == ShutdownRepositoryLocationStatus.SUCCESS, result.message

        # Wait for client to be unavailable
        start_time = time.time()

        while time.time() - start_time < 15:
            try:
                origin.create_client().heartbeat()
            except DagsterUserCodeUnreachableError:
                # Shutdown succeeded
                return
            time.sleep(1)

        raise Exception("Timed out waiting for shutdown to take effect")

    def test_shutdown_repository_location_not_found(self, graphql_client):
        result = graphql_client.shutdown_repository_location("not_real")

        assert result.status == ShutdownRepositoryLocationStatus.FAILURE
        assert "Location not_real does not exist" in result.message



from collections.abc import Sequence
from typing import TYPE_CHECKING, AbstractSet, Optional  # noqa: UP035

import dagster._check as check
import graphene
from dagster._core.definitions.asset_health.asset_freshness_health import AssetFreshnessHealthState
from dagster._core.definitions.asset_health.asset_materialization_health import (
    MinimalAssetMaterializationHealthState,
)
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.freshness import FreshnessStateRecord
from dagster._core.definitions.partitions.utils import PartitionRangeStatus
from dagster._core.errors import DagsterUserCodeProcessError
from dagster._core.event_api import EventLogCursor
from dagster._core.events import DagsterEventType
from dagster._core.remote_representation.code_location import is_implicit_asset_job_name
from dagster._core.remote_representation.external import RemoteExecutionPlan, RemoteJob
from dagster._core.remote_representation.external_data import (
    DEFAULT_MODE_NAME,
    PartitionExecutionErrorSnap,
    PresetSnap,
)
from dagster._core.remote_representation.represented import RepresentedJob
from dagster._core.storage.dagster_run import (
    DagsterRunStatsSnapshot,
    DagsterRunStatus,
    RunRecord,
    RunsFilter,
)
from dagster._core.storage.event_log.base import AssetRecord
from dagster._core.storage.tags import (
    EXTERNAL_JOB_SOURCE_TAG_KEY,
    REPOSITORY_LABEL_TAG,
    RUN_METRIC_TAGS,
    TagType,
    get_tag_type,
)
from dagster._core.workspace.permissions import Permissions
from dagster._utils.tags import get_boolean_tag_value
from dagster_shared.yaml_utils import dump_run_config_yaml

from dagster_graphql.implementation.events import (
    get_graphene_events_from_records_connection,
    iterate_metadata_entries,
)
from dagster_graphql.implementation.fetch_asset_checks import get_asset_checks_for_run_id
from dagster_graphql.implementation.fetch_assets import get_assets_for_run
from dagster_graphql.implementation.fetch_pipelines import get_job_reference_or_raise
from dagster_graphql.implementation.fetch_runs import get_runs, get_stats, get_step_stats
from dagster_graphql.implementation.fetch_schedules import get_schedules_for_job
from dagster_graphql.implementation.fetch_sensors import get_sensors_for_job
from dagster_graphql.implementation.loader import RepositoryScopedBatchLoader
from dagster_graphql.implementation.utils import (
    UserFacingGraphQLError,
    apply_cursor_limit_reverse,
    capture_error,
    get_query_limit_with_default,
    has_permission_for_definition,
    has_permission_for_run,
)
from dagster_graphql.schema.asset_health import GrapheneAssetHealth
from dagster_graphql.schema.dagster_types import (
    GrapheneDagsterType,
    GrapheneDagsterTypeOrError,
    GrapheneDagsterTypeUnion,
    to_dagster_type,
)
from dagster_graphql.schema.entity_key import GrapheneAssetCheckHandle, GrapheneAssetKey
from dagster_graphql.schema.errors import (
    GrapheneDagsterTypeNotFoundError,
    GraphenePythonError,
    GrapheneRunNotFoundError,
)
from dagster_graphql.schema.execution import GrapheneExecutionPlan
from dagster_graphql.schema.inputs import GrapheneAssetKeyInput
from dagster_graphql.schema.logs.compute_logs import GrapheneCapturedLogs, from_captured_log_data
from dagster_graphql.schema.logs.events import (
    GrapheneAssetMaterializationEventType,
    GrapheneAssetResultEventType,
    GrapheneDagsterRunEvent,
    GrapheneFailedToMaterializeEvent,
    GrapheneMaterializationEvent,
    GrapheneObservationEvent,
    GrapheneRunStepStats,
)
from dagster_graphql.schema.metadata import GrapheneMetadataEntry
from dagster_graphql.schema.owners import GrapheneDefinitionOwner, definition_owner_from_owner_str
from dagster_graphql.schema.partition_keys import GraphenePartitionKeys
from dagster_graphql.schema.pipelines.mode import GrapheneMode
from dagster_graphql.schema.pipelines.pipeline_ref import GraphenePipelineReference
from dagster_graphql.schema.pipelines.pipeline_run_stats import GrapheneRunStatsSnapshotOrError
from dagster_graphql.schema.pipelines.status import GrapheneRunStatus
from dagster_graphql.schema.repository_origin import GrapheneRepositoryOrigin
from dagster_graphql.schema.runs import GrapheneRunConfigData
from dagster_graphql.schema.runs_feed import GrapheneRunsFeedEntry
from dagster_graphql.schema.schedules.schedules import GrapheneSchedule
from dagster_graphql.schema.sensors import GrapheneSensor
from dagster_graphql.schema.solids import (
    GrapheneSolid,
    GrapheneSolidContainer,
    GrapheneSolidHandle,
    build_solid_handles,
    build_solids,
)
from dagster_graphql.schema.tags import GraphenePipelineTag
from dagster_graphql.schema.util import ResolveInfo, get_compute_log_manager, non_null_list

if TYPE_CHECKING:
    from dagster_graphql.schema.asset_graph import GrapheneAssetNode
    from dagster_graphql.schema.partition_sets import GrapheneJobSelectionPartition

UNSTARTED_STATUSES = [
    DagsterRunStatus.QUEUED,
    DagsterRunStatus.NOT_STARTED,
    DagsterRunStatus.STARTING,
]

STARTED_STATUSES = {
    DagsterRunStatus.STARTED,
    DagsterRunStatus.SUCCESS,
    DagsterRunStatus.FAILURE,
    DagsterRunStatus.CANCELED,
}

COMPLETED_STATUSES = {
    DagsterRunStatus.FAILURE,
    DagsterRunStatus.SUCCESS,
    DagsterRunStatus.CANCELED,
}


def parse_timestamp(timestamp: str | None = None) -> float | None:
    try:
        return int(timestamp) / 1000.0 if timestamp else None
    except ValueError:
        return None


GraphenePartitionRangeStatus = graphene.Enum.from_enum(PartitionRangeStatus)


class GrapheneTimePartitionRange(graphene.ObjectType):
    startTime = graphene.NonNull(graphene.Float)
    endTime = graphene.NonNull(graphene.Float)
    startKey = graphene.NonNull(graphene.String)
    endKey = graphene.NonNull(graphene.String)

    class Meta:
        name = "TimePartitionRange"


class GrapheneTimePartitionRangeStatus(GrapheneTimePartitionRange):
    status = graphene.NonNull(GraphenePartitionRangeStatus)

    class Meta:  # pyright: ignore[reportIncompatibleVariableOverride]
        name = "TimePartitionRangeStatus"


class GrapheneTimePartitionStatuses(graphene.ObjectType):
    ranges = non_null_list(GrapheneTimePartitionRangeStatus)

    class Meta:
        name = "TimePartitionStatuses"


class GrapheneDefaultPartitionStatuses(graphene.ObjectType):
    materializedPartitions = non_null_list(graphene.String)
    failedPartitions = non_null_list(graphene.String)
    unmaterializedPartitions = non_null_list(graphene.String)
    materializingPartitions = non_null_list(graphene.String)

    class Meta:
        name = "DefaultPartitionStatuses"


class GraphenePartitionStatus1D(graphene.Union):
    class Meta:
        types = (GrapheneTimePartitionStatuses, GrapheneDefaultPartitionStatuses)
        name = "PartitionStatus1D"


class GrapheneMultiPartitionRangeStatuses(graphene.ObjectType):
    """The primary dimension of a multipartitioned asset is the time-partitioned dimension.
    If both dimensions of the asset are static or time-partitioned, the primary dimension is
    the first defined dimension.
    """

    primaryDimStartKey = graphene.NonNull(graphene.String)
    primaryDimEndKey = graphene.NonNull(graphene.String)
    primaryDimStartTime = graphene.Field(graphene.Float)
    primaryDimEndTime = graphene.Field(graphene.Float)
    secondaryDim = graphene.NonNull(GraphenePartitionStatus1D)

    class Meta:
        name = "MaterializedPartitionRangeStatuses2D"


class GrapheneMultiPartitionStatuses(graphene.ObjectType):
    ranges = non_null_list(GrapheneMultiPartitionRangeStatuses)
    primaryDimensionName = graphene.NonNull(graphene.String)

    class Meta:
        name = "MultiPartitionStatuses"


class GrapheneAssetPartitionStatuses(graphene.Union):
    class Meta:
        types = (
            GrapheneDefaultPartitionStatuses,
            GrapheneMultiPartitionStatuses,
            GrapheneTimePartitionStatuses,
        )
        name = "AssetPartitionStatuses"


class GraphenePartitionStats(graphene.ObjectType):
    numMaterialized = graphene.NonNull(graphene.Int)
    numPartitions = graphene.NonNull(graphene.Int)
    numFailed = graphene.NonNull(graphene.Int)
    numMaterializing = graphene.NonNull(graphene.Int)

    class Meta:
        name = "PartitionStats"


class GrapheneMaterializationHistoryEventTypeSelector(graphene.Enum):
    MATERIALIZATION = "MATERIALIZATION"
    FAILED_TO_MATERIALIZE = "FAILED_TO_MATERIALIZE"
    ALL = "ALL"

    class Meta:
        name = "MaterializationHistoryEventTypeSelector"


class GrapheneAssetEventHistoryEventTypeSelector(graphene.Enum):
    MATERIALIZATION = "MATERIALIZATION"
    FAILED_TO_MATERIALIZE = "FAILED_TO_MATERIALIZE"
    OBSERVATION = "OBSERVATION"

    class Meta:
        name = "AssetEventHistoryEventTypeSelector"


class GrapheneMaterializationHistoryConnection(graphene.ObjectType):
    class Meta:
        name = "MaterializationHistoryConnection"

    results = non_null_list(GrapheneAssetMaterializationEventType)
    cursor = graphene.NonNull(graphene.String)


class GrapheneAssetResultEventHistoryConnection(graphene.ObjectType):
    class Meta:
        name = "AssetResultEventHistoryConnection"

    results = non_null_list(GrapheneAssetResultEventType)
    cursor = graphene.NonNull(graphene.String)


class GrapheneAsset(graphene.ObjectType):
    id = graphene.NonNull(graphene.String)
    key = graphene.NonNull(GrapheneAssetKey)
    assetMaterializations = graphene.Field(
        non_null_list(GrapheneMaterializationEvent),
        partitions=graphene.List(graphene.NonNull(graphene.String)),
        beforeTimestampMillis=graphene.String(),
        afterTimestampMillis=graphene.String(),
        limit=graphene.Int(),
    )
    assetObservations = graphene.Field(
        non_null_list(GrapheneObservationEvent),
        partitions=graphene.List(graphene.NonNull(graphene.String)),
        beforeTimestampMillis=graphene.String(),
        afterTimestampMillis=graphene.String(),
        limit=graphene.Int(),
    )
    assetEventHistory = graphene.Field(
        graphene.NonNull(GrapheneAssetResultEventHistoryConnection),
        partitions=graphene.List(graphene.NonNull(graphene.String)),
        beforeTimestampMillis=graphene.String(),
        afterTimestampMillis=graphene.String(),
        limit=graphene.NonNull(graphene.Int),
        eventTypeSelectors=graphene.NonNull(
            graphene.List(graphene.NonNull(GrapheneAssetEventHistoryEventTypeSelector))
        ),
        cursor=graphene.String(),
    )

    definition = graphene.Field("dagster_graphql.schema.asset_graph.GrapheneAssetNode")
    latestEventSortKey = graphene.Field(graphene.ID)
    assetHealth = graphene.Field(GrapheneAssetHealth)
    latestMaterializationTimestamp = graphene.Float()
    hasDefinitionOrRecord = graphene.NonNull(graphene.Boolean)
    latestFailedToMaterializeTimestamp = graphene.Float()
    freshnessStatusChangedTimestamp = graphene.Float()

    class Meta:
        name = "Asset"

    def __init__(self, key):
        self._asset_key = key
        super().__init__(
            key=GrapheneAssetKey(path=key.path),
        )

    def resolve_id(self, _) -> str:
        return self._asset_key.to_string()

    def resolve_definition(self, graphene_info: ResolveInfo) -> Optional["GrapheneAssetNode"]:
        from dagster_graphql.schema.asset_graph import GrapheneAssetNode

        remote_asset_node = (
            graphene_info.context.asset_graph.get(self._asset_key)
            if graphene_info.context.asset_graph.has(self._asset_key)
            else None
        )
        return GrapheneAssetNode(remote_asset_node) if remote_asset_node else None

    async def resolve_assetMaterializations(
        self,
        graphene_info: ResolveInfo,
        partitions: Sequence[str] | None = None,
        beforeTimestampMillis: str | None = None,
        afterTimestampMillis: str | None = None,
        limit: int | None = None,
    ) -> Sequence[GrapheneMaterializationEvent]:
        from dagster_graphql.implementation.fetch_assets import get_asset_materializations

        before_timestamp = parse_timestamp(beforeTimestampMillis)
        after_timestamp = parse_timestamp(afterTimestampMillis)

        if limit == 1 and not partitions and not before_timestamp and not after_timestamp:
            record = await AssetRecord.gen(graphene_info.context, self._asset_key)
            latest_materialization_event = (
                record.asset_entry.last_materialization if record else None
            )

            if not latest_materialization_event:
                return []

            return [GrapheneMaterializationEvent(event=latest_materialization_event)]

        events = get_asset_materializations(
            graphene_info,
            self._asset_key,
            partitions=partitions,
            before_timestamp=before_timestamp,
            after_timestamp=after_timestamp,
            limit=limit,
        )
        return [GrapheneMaterializationEvent(event=event) for event in events]

    def resolve_assetEventHistory(
        self,
        graphene_info: ResolveInfo,
        eventTypeSelectors: Sequence[GrapheneAssetEventHistoryEventTypeSelector],
        limit: int | None,
        partitions: Sequence[str] | None = None,
        beforeTimestampMillis: str | None = None,
        afterTimestampMillis: str | None = None,
        cursor: str | None = None,
    ) -> GrapheneAssetResultEventHistoryConnection:
        from dagster_graphql.implementation.fetch_assets import (
            get_asset_failed_to_materialize_event_records,
            get_asset_materialization_event_records,
            get_asset_observation_event_records,
        )

        before_timestamp = parse_timestamp(beforeTimestampMillis)
        after_timestamp = parse_timestamp(afterTimestampMillis)
        failure_events = []
        success_events = []
        observation_events = []

        if GrapheneAssetEventHistoryEventTypeSelector.FAILED_TO_MATERIALIZE in eventTypeSelectors:
            failure_events = [
                (record.storage_id, GrapheneFailedToMaterializeEvent(event=record.event_log_entry))
                for record in get_asset_failed_to_materialize_event_records(
                    graphene_info,
                    self._asset_key,
                    partitions=partitions,
                    before_timestamp=before_timestamp,
                    after_timestamp=after_timestamp,
                    limit=limit,
                    cursor=cursor,
                )
            ]

        if GrapheneAssetEventHistoryEventTypeSelector.MATERIALIZATION in eventTypeSelectors:
            success_events = [
                (record.storage_id, GrapheneMaterializationEvent(event=record.event_log_entry))
                for record in get_asset_materialization_event_records(
                    graphene_info,
                    self._asset_key,
                    partitions=partitions,
                    before_timestamp=before_timestamp,
                    after_timestamp=after_timestamp,
                    limit=limit,
                    cursor=cursor,
                )
            ]

        if GrapheneAssetEventHistoryEventTypeSelector.OBSERVATION in eventTypeSelectors:
            observation_events = [
                (record.storage_id, GrapheneObservationEvent(event=record.event_log_entry))
                for record in get_asset_observation_event_records(
                    graphene_info,
                    self._asset_key,
                    partitions=partitions,
                    before_timestamp=before_timestamp,
                    after_timestamp=after_timestamp,
                    limit=limit,
                    cursor=cursor,
                )
            ]

        all_events_tuples = failure_events + success_events + observation_events
        sorted_limited_event_tuples = sorted(
            all_events_tuples, key=lambda event_tuple: event_tuple[0], reverse=True
        )[:limit]

        new_cursor = (
            EventLogCursor.from_storage_id(sorted_limited_event_tuples[-1][0]).to_string()
            if sorted_limited_event_tuples
            else EventLogCursor.from_storage_id(-1).to_string()
        )

        return GrapheneAssetResultEventHistoryConnection(
            results=[event for _, event in sorted_limited_event_tuples],
            cursor=new_cursor,
        )

    def resolve_assetObservations(
        self,
        graphene_info: ResolveInfo,
        partitions: Sequence[str] | None = None,
        beforeTimestampMillis: str | None = None,
        afterTimestampMillis: str | None = None,
        limit: int | None = None,
    ) -> Sequence[GrapheneObservationEvent]:
        from dagster_graphql.implementation.fetch_assets import get_asset_observations

        before_timestamp = parse_timestamp(beforeTimestampMillis)
        after_timestamp = parse_timestamp(afterTimestampMillis)

        return [
            GrapheneObservationEvent(event=event)
            for event in get_asset_observations(
                graphene_info,
                self._asset_key,
                partitions=partitions,
                before_timestamp=before_timestamp,
                after_timestamp=after_timestamp,
                limit=limit,
            )
        ]

    async def resolve_latestEventSortKey(self, graphene_info):
        asset_record = await AssetRecord.gen(graphene_info.context, self._asset_key)
        if asset_record:
            return asset_record.asset_entry.last_event_storage_id
        return None

    def resolve_assetHealth(self, graphene_info: ResolveInfo) -> GrapheneAssetHealth | None:
        if not graphene_info.context.instance.dagster_asset_health_queries_supported():
            return None
        return GrapheneAssetHealth(
            asset_key=self._asset_key,
            dynamic_partitions_loader=graphene_info.context.dynamic_partitions_loader,
        )

    async def resolve_hasDefinitionOrRecord(self, graphene_info: ResolveInfo) -> bool:
        return (
            graphene_info.context.asset_graph.has(self._asset_key)
            or await AssetRecord.gen(graphene_info.context, self._asset_key) is not None
        )

    async def resolve_latestMaterializationTimestamp(
        self, graphene_info: ResolveInfo
    ) -> float | None:
        min_materialization_state = await MinimalAssetMaterializationHealthState.gen(
            graphene_info.context, self._asset_key
        )
        if min_materialization_state is not None:
            return (
                min_materialization_state.latest_materialization_timestamp
                * 1000  # FE prefers timestamp in milliseconds
                if min_materialization_state.latest_materialization_timestamp
                else None
            )

        record = await AssetRecord.gen(graphene_info.context, self._asset_key)
        latest_materialization_event = record.asset_entry.last_materialization if record else None
        return (
            latest_materialization_event.timestamp * 1000  # FE prefers timestamp in milliseconds
            if latest_materialization_event
            else None
        )

    async def resolve_latestFailedToMaterializeTimestamp(
        self, graphene_info: ResolveInfo
    ) -> float | None:
        materialization_state = await MinimalAssetMaterializationHealthState.gen(
            graphene_info.context, self._asset_key
        )
        if materialization_state is not None:
            ts = materialization_state.latest_failed_to_materialize_timestamp
        else:
            record = await AssetRecord.gen(graphene_info.context, self._asset_key)
            latest_failed_to_materialize_event = (
                record.asset_entry.last_failed_to_materialize_entry if record else None
            )
            ts = (
                latest_failed_to_materialize_event.timestamp
                if latest_failed_to_materialize_event
                else None
            )

        return ts * 1000 if ts else None  # FE prefers timestamp in milliseconds

    async def resolve_freshnessStatusChangedTimestamp(
        self, graphene_info: ResolveInfo
    ) -> float | None:
        freshness_state = await AssetFreshnessHealthState.gen(
            graphene_info.context, self._asset_key
        )
        if freshness_state is not None:
            ts = freshness_state.updated_timestamp
        else:
            freshness_state_record = await FreshnessStateRecord.gen(
                graphene_info.context, self._asset_key
            )

            ts = freshness_state_record.updated_at.timestamp() if freshness_state_record else None

        return ts * 1000 if ts else None  # FE prefers timestamp in milliseconds


class GrapheneEventConnection(graphene.ObjectType):
    class Meta:
        name = "EventConnection"

    events = non_null_list(GrapheneDagsterRunEvent)
    cursor = graphene.NonNull(graphene.String)
    hasMore = graphene.NonNull(graphene.Boolean)


class GrapheneEventConnectionOrError(graphene.Union):
    class Meta:
        types = (GrapheneEventConnection, GrapheneRunNotFoundError, GraphenePythonError)
        name = "EventConnectionOrError"


class GraphenePipelineRun(graphene.Interface):
    id = graphene.NonNull(graphene.ID)
    runId = graphene.NonNull(graphene.String)
    # Nullable because of historical runs
    pipelineSnapshotId = graphene.String()
    repositoryOrigin = graphene.Field(GrapheneRepositoryOrigin)
    status = graphene.NonNull(GrapheneRunStatus)
    pipeline = graphene.NonNull(GraphenePipelineReference)
    pipelineName = graphene.NonNull(graphene.String)
    jobName = graphene.NonNull(graphene.String)
    solidSelection = graphene.List(graphene.NonNull(graphene.String))
    stats = graphene.NonNull(GrapheneRunStatsSnapshotOrError)
    stepStats = non_null_list(GrapheneRunStepStats)
    capturedLogs = graphene.Field(
        graphene.NonNull(GrapheneCapturedLogs),
        fileKey=graphene.Argument(graphene.NonNull(graphene.String)),
        description="""
        Captured logs are the stdout/stderr logs for a given file key within the run
        """,
    )
    executionPlan = graphene.Field(GrapheneExecutionPlan)
    stepKeysToExecute = graphene.List(graphene.NonNull(graphene.String))
    runConfigYaml = graphene.NonNull(graphene.String)
    runConfig = graphene.NonNull(GrapheneRunConfigData)
    mode = graphene.NonNull(graphene.String)
    tags = non_null_list(GraphenePipelineTag)
    rootRunId = graphene.Field(graphene.String)
    parentRunId = graphene.Field(graphene.String)
    canTerminate = graphene.NonNull(graphene.Boolean)
    assets = non_null_list(GrapheneAsset)
    eventConnection = graphene.Field(
        graphene.NonNull(GrapheneEventConnection),
        afterCursor=graphene.Argument(graphene.String),
        limit=graphene.Argument(graphene.Int),
    )

    class Meta:
        name = "PipelineRun"


class GrapheneRun(graphene.ObjectType):
    id = graphene.NonNull(graphene.ID)
    runId = graphene.NonNull(graphene.String)
    # Nullable because of historical runs
    pipelineSnapshotId = graphene.String()
    parentPipelineSnapshotId = graphene.String()
    repositoryOrigin = graphene.Field(GrapheneRepositoryOrigin)
    status = graphene.NonNull(GrapheneRunStatus)
    runStatus = graphene.Field(
        graphene.NonNull(GrapheneRunStatus),
        description="Included to comply with RunsFeedEntry interface. Duplicate of status.",
    )
    pipeline = graphene.NonNull(GraphenePipelineReference)
    pipelineName = graphene.NonNull(graphene.String)
    jobName = graphene.NonNull(graphene.String)
    solidSelection = graphene.List(graphene.NonNull(graphene.String))
    assetSelection = graphene.List(graphene.NonNull(GrapheneAssetKey))
    assetCheckSelection = graphene.List(graphene.NonNull(GrapheneAssetCheckHandle))
    resolvedOpSelection = graphene.List(graphene.NonNull(graphene.String))
    stats = graphene.NonNull(GrapheneRunStatsSnapshotOrError)
    stepStats = non_null_list(GrapheneRunStepStats)
    executionPlan = graphene.Field(GrapheneExecutionPlan)
    stepKeysToExecute = graphene.List(graphene.NonNull(graphene.String))
    runConfigYaml = graphene.NonNull(graphene.String)
    runConfig = graphene.NonNull(GrapheneRunConfigData)
    mode = graphene.NonNull(graphene.String)
    tags = non_null_list(GraphenePipelineTag)
    rootRunId = graphene.Field(graphene.String)
    parentRunId = graphene.Field(graphene.String)
    canTerminate = graphene.NonNull(graphene.Boolean)
    assetMaterializations = non_null_list(GrapheneMaterializationEvent)
    assets = non_null_list(GrapheneAsset)
    assetChecks = graphene.List(graphene.NonNull(GrapheneAssetCheckHandle))
    eventConnection = graphene.Field(
        graphene.NonNull(GrapheneEventConnection),
        afterCursor=graphene.Argument(graphene.String),
        limit=graphene.Argument(graphene.Int),
    )
    creationTime = graphene.NonNull(graphene.Float)
    startTime = graphene.Float()
    endTime = graphene.Float()
    updateTime = graphene.Float()
    hasReExecutePermission = graphene.NonNull(graphene.Boolean)
    hasTerminatePermission = graphene.NonNull(graphene.Boolean)
    hasDeletePermission = graphene.NonNull(graphene.Boolean)
    hasConcurrencyKeySlots = graphene.NonNull(graphene.Boolean)
    rootConcurrencyKeys = graphene.List(graphene.NonNull(graphene.String))
    allPools = graphene.List(graphene.NonNull(graphene.String))
    hasUnconstrainedRootNodes = graphene.NonNull(graphene.Boolean)
    hasRunMetricsEnabled = graphene.NonNull(graphene.Boolean)
    externalJobSource = graphene.String()

    class Meta:
        interfaces = (GraphenePipelineRun, GrapheneRunsFeedEntry)
        name = "Run"

    def __init__(self, record: RunRecord):
        check.inst_param(record, "record", RunRecord)
        dagster_run = record.dagster_run
        super().__init__(
            runId=dagster_run.run_id,
            status=dagster_run.status.value,
            runStatus=dagster_run.status.value,
            mode=DEFAULT_MODE_NAME,
        )
        self.dagster_run = dagster_run
        self._run_record = record
        self._run_stats: DagsterRunStatsSnapshot | None = None

    @property
    def creation_timestamp(self) -> float:
        return self._run_record.create_timestamp.timestamp()

    def resolve_hasReExecutePermission(self, graphene_info: ResolveInfo):
        return has_permission_for_run(
            graphene_info, Permissions.LAUNCH_PIPELINE_REEXECUTION, self.dagster_run
        )

    def resolve_hasTerminatePermission(self, graphene_info: ResolveInfo):
        return has_permission_for_run(
            graphene_info, Permissions.TERMINATE_PIPELINE_EXECUTION, self.dagster_run
        )

    def resolve_hasDeletePermission(self, graphene_info: ResolveInfo):
        return has_permission_for_run(
            graphene_info, Permissions.DELETE_PIPELINE_RUN, self.dagster_run
        )

    def resolve_id(self, _graphene_info: ResolveInfo):
        return self.dagster_run.run_id

    def resolve_repositoryOrigin(self, _graphene_info: ResolveInfo):
        return (
            GrapheneRepositoryOrigin(self.dagster_run.remote_job_origin.repository_origin)
            if self.dagster_run.remote_job_origin
            else None
        )

    def resolve_pipeline(self, graphene_info: ResolveInfo):
        return get_job_reference_or_raise(graphene_info, self.dagster_run)

    def resolve_pipelineName(self, _graphene_info: ResolveInfo):
        return self.dagster_run.job_name

    def resolve_jobName(self, _graphene_info: ResolveInfo):
        return self.dagster_run.job_name

    def resolve_solidSelection(self, _graphene_info: ResolveInfo):
        return self.dagster_run.op_selection

    def resolve_assetSelection(self, _graphene_info: ResolveInfo):
        return self.dagster_run.asset_selection

    def resolve_assetCheckSelection(self, _graphene_info: ResolveInfo):
        return (
            [GrapheneAssetCheckHandle(handle) for handle in self.dagster_run.asset_check_selection]
            if self.dagster_run.asset_check_selection is not None
            else None
        )

    def resolve_resolvedOpSelection(self, _graphene_info: ResolveInfo):
        return self.dagster_run.resolved_op_selection

    def resolve_pipelineSnapshotId(self, _graphene_info: ResolveInfo):
        return self.dagster_run.job_snapshot_id

    def resolve_parentPipelineSnapshotId(self, graphene_info: ResolveInfo):
        pipeline_snapshot_id = self.dagster_run.job_snapshot_id
        if pipeline_snapshot_id is not None and graphene_info.context.instance.has_job_snapshot(
            pipeline_snapshot_id
        ):
            snapshot = graphene_info.context.instance.get_job_snapshot(pipeline_snapshot_id)
            if snapshot.lineage_snapshot is not None:
                return snapshot.lineage_snapshot.parent_snapshot_id
        return None

    @capture_error
    def resolve_stats(self, graphene_info: ResolveInfo):
        return get_stats(graphene_info, self.run_id)

    def resolve_stepStats(self, graphene_info: ResolveInfo):
        return get_step_stats(graphene_info, self.run_id)

    def resolve_capturedLogs(self, graphene_info: ResolveInfo, fileKey):
        compute_log_manager = get_compute_log_manager(graphene_info)
        log_key = compute_log_manager.build_log_key_for_run(self.run_id, fileKey)
        log_data = compute_log_manager.get_log_data(log_key)
        return from_captured_log_data(log_data)

    def resolve_executionPlan(self, graphene_info: ResolveInfo):
        if not (self.dagster_run.execution_plan_snapshot_id and self.dagster_run.job_snapshot_id):
            return None

        instance = graphene_info.context.instance

        execution_plan_snapshot = instance.get_execution_plan_snapshot(
            self.dagster_run.execution_plan_snapshot_id
        )
        return (
            GrapheneExecutionPlan(
                RemoteExecutionPlan(execution_plan_snapshot=execution_plan_snapshot)
            )
            if execution_plan_snapshot
            else None
        )

    def resolve_stepKeysToExecute(self, _graphene_info: ResolveInfo):
        return self.dagster_run.step_keys_to_execute

    def resolve_runConfigYaml(self, _graphene_info: ResolveInfo):
        return dump_run_config_yaml(self.dagster_run.run_config)

    def resolve_runConfig(self, _graphene_info: ResolveInfo):
        return self.dagster_run.run_config

    def resolve_tags(self, _graphene_info: ResolveInfo):
        return [
            GraphenePipelineTag(key=key, value=value)
            for key, value in self.dagster_run.tags.items()
            if get_tag_type(key) != TagType.HIDDEN
        ]

    def resolve_externalJobSource(self, _graphene_info: ResolveInfo):
        source_str = self.dagster_run.tags.get(EXTERNAL_JOB_SOURCE_TAG_KEY)
        if source_str:
            return source_str.lower()
        return None

    def resolve_rootRunId(self, _graphene_info: ResolveInfo):
        return self.dagster_run.root_run_id

    def resolve_parentRunId(self, _graphene_info: ResolveInfo):
        return self.dagster_run.parent_run_id

    @property
    def run_id(self):
        return self.runId

    def resolve_canTerminate(self, _graphene_info: ResolveInfo):
        # short circuit if the pipeline run is in a terminal state
        if self.dagster_run.is_finished:
            return False
        return (
            self.dagster_run.status == DagsterRunStatus.QUEUED
            or self.dagster_run.status == DagsterRunStatus.STARTED
        )

    def resolve_assets(self, graphene_info: ResolveInfo):
        return get_assets_for_run(graphene_info, self.dagster_run)

    def resolve_assetChecks(self, graphene_info: ResolveInfo):
        return get_asset_checks_for_run_id(graphene_info, self.run_id)

    def resolve_assetMaterializations(self, graphene_info: ResolveInfo):
        # convenience field added for users querying directly via GraphQL
        return [
            GrapheneMaterializationEvent(event=event)
            for event in graphene_info.context.instance.all_logs(
                self.run_id, of_type=DagsterEventType.ASSET_MATERIALIZATION
            )
        ]

    def resolve_eventConnection(self, graphene_info: ResolveInfo, afterCursor=None, limit=None):
        default_limit = graphene_info.context.records_for_run_default_limit
        if default_limit:
            limit = get_query_limit_with_default(limit, default_limit)

        conn = graphene_info.context.instance.get_records_for_run(
            self.run_id, cursor=afterCursor, limit=limit
        )
        return GrapheneEventConnection(
            events=get_graphene_events_from_records_connection(
                graphene_info.context.instance, conn, self.dagster_run.job_name
            ),
            cursor=conn.cursor,
            hasMore=conn.has_more,
        )

    def resolve_startTime(self, graphene_info: ResolveInfo):
        # If a user has not migrated in 0.13.15, then run_record will not have start_time and end_time. So it will be necessary to fill this data using the run_stats. Since we potentially make this call multiple times, we cache the result.
        if self._run_record.start_time is None and self.dagster_run.status in STARTED_STATUSES:
            # Short-circuit if pipeline failed to start, so it has an end time but no start time
            if self._run_record.end_time is not None:
                return self._run_record.end_time

            if self._run_stats is None or self._run_stats.start_time is None:
                self._run_stats = graphene_info.context.instance.get_run_stats(self.runId)

            if self._run_stats.start_time is None and self._run_stats.end_time:
                return self._run_stats.end_time

            return self._run_stats.start_time
        return self._run_record.start_time

    def resolve_endTime(self, graphene_info: ResolveInfo):
        if self._run_record.end_time is None and self.dagster_run.status in COMPLETED_STATUSES:
            if self._run_stats is None or self._run_stats.end_time is None:
                self._run_stats = graphene_info.context.instance.get_run_stats(self.runId)
            return self._run_stats.end_time
        return self._run_record.end_time

    def resolve_updateTime(self, graphene_info: ResolveInfo):
        return self._run_record.update_timestamp.timestamp()

    def resolve_creationTime(self, graphene_info: ResolveInfo):
        return self.creation_timestamp

    def resolve_hasConcurrencyKeySlots(self, graphene_info: ResolveInfo):
        instance = graphene_info.context.instance
        if not instance.event_log_storage.supports_global_concurrency_limits:
            return False

        active_run_ids = instance.event_log_storage.get_concurrency_run_ids()
        return self.runId in active_run_ids

    def resolve_hasUnconstrainedRootNodes(self, graphene_info: ResolveInfo):
        if not self.dagster_run.run_op_concurrency:
            return True

        if self.dagster_run.run_op_concurrency.has_unconstrained_root_nodes:
            return True

        return False

    def resolve_allPools(self, graphene_info: ResolveInfo):
        if not self.dagster_run.run_op_concurrency:
            return None

        return (
            list(self.dagster_run.run_op_concurrency.all_pools)
            if self.dagster_run.run_op_concurrency.all_pools
            else []
        )

    def resolve_rootConcurrencyKeys(self, graphene_info: ResolveInfo):
        if not self.dagster_run.run_op_concurrency:
            return None

        root_concurrency_keys = []
        for concurrency_key, count in self.dagster_run.run_op_concurrency.root_key_counts.items():
            root_concurrency_keys.extend([concurrency_key] * count)
        return root_concurrency_keys

    def resolve_hasRunMetricsEnabled(self, graphene_info: ResolveInfo):
        if self.dagster_run.status in UNSTARTED_STATUSES:
            return False

        run_tags = self.dagster_run.tags
        return any(get_boolean_tag_value(run_tags.get(tag)) for tag in RUN_METRIC_TAGS)


class GrapheneIPipelineSnapshotMixin:
    # Mixin this class to implement IPipelineSnapshot
    #
    # Graphene has some strange properties that make it so that you cannot
    # implement ABCs nor use properties in an overridable way. So the way
    # the mixin works is that the target classes have to have a method
    # get_represented_job()
    #
    name = graphene.NonNull(graphene.String)
    description = graphene.String()
    owners = non_null_list(GrapheneDefinitionOwner)
    id = graphene.NonNull(graphene.ID)
    pipeline_snapshot_id = graphene.NonNull(graphene.String)
    dagster_types = non_null_list(GrapheneDagsterType)
    dagster_type_or_error = graphene.Field(
        graphene.NonNull(GrapheneDagsterTypeOrError),
        dagsterTypeName=graphene.Argument(graphene.NonNull(graphene.String)),
    )
    solids = non_null_list(GrapheneSolid)
    modes = non_null_list(GrapheneMode)
    solid_handles = graphene.Field(
        non_null_list(GrapheneSolidHandle), parentHandleID=graphene.String()
    )
    solid_handle = graphene.Field(
        GrapheneSolidHandle,
        handleID=graphene.Argument(graphene.NonNull(graphene.String)),
    )
    tags = non_null_list(GraphenePipelineTag)
    run_tags = non_null_list(GraphenePipelineTag)
    metadata_entries = non_null_list(GrapheneMetadataEntry)
    runs = graphene.Field(
        non_null_list(GrapheneRun),
        cursor=graphene.String(),
        limit=graphene.Int(),
    )
    schedules = non_null_list(GrapheneSchedule)
    sensors = non_null_list(GrapheneSensor)
    parent_snapshot_id = graphene.String()
    graph_name = graphene.NonNull(graphene.String)
    externalJobSource = graphene.String()

    class Meta:
        name = "IPipelineSnapshotMixin"

    def get_represented_job(self) -> RepresentedJob:
        raise NotImplementedError()

    def resolve_pipeline_snapshot_id(self, _graphene_info: ResolveInfo):
        return self.get_represented_job().identifying_job_snapshot_id

    def resolve_id(self, _graphene_info: ResolveInfo):
        return self.get_represented_job().identifying_job_snapshot_id

    def resolve_name(self, _graphene_info: ResolveInfo):
        return self.get_represented_job().name

    def resolve_description(self, _graphene_info: ResolveInfo):
        return self.get_represented_job().description

    def resolve_owners(self, _graphene_info: ResolveInfo):
        return [
            definition_owner_from_owner_str(owner)
            for owner in (self.get_represented_job().owners or [])
        ]

    def resolve_dagster_types(self, _graphene_info: ResolveInfo):
        represented_pipeline = self.get_represented_job()
        return sorted(
            list(
                map(
                    lambda dt: to_dagster_type(
                        represented_pipeline.job_snapshot.dagster_type_namespace_snapshot.get_dagster_type_snap,
                        represented_pipeline.job_snapshot.config_schema_snapshot.get_config_snap,
                        dt.key,
                    ),
                    [t for t in represented_pipeline.dagster_type_snaps if t.name],
                )
            ),
            key=lambda dagster_type: dagster_type.name,
        )

    @capture_error
    def resolve_dagster_type_or_error(
        self, _graphene_info: ResolveInfo, dagsterTypeName: str
    ) -> GrapheneDagsterTypeUnion:
        represented_pipeline = self.get_represented_job()

        if not represented_pipeline.has_dagster_type_named(dagsterTypeName):
            raise UserFacingGraphQLError(
                GrapheneDagsterTypeNotFoundError(dagster_type_name=dagsterTypeName)
            )

        return to_dagster_type(
            represented_pipeline.job_snapshot.dagster_type_namespace_snapshot.get_dagster_type_snap,
            represented_pipeline.job_snapshot.config_schema_snapshot.get_config_snap,
            represented_pipeline.get_dagster_type_by_name(dagsterTypeName).key,
        )

    def resolve_solids(self, _graphene_info: ResolveInfo):
        represented_pipeline = self.get_represented_job()
        return build_solids(
            represented_pipeline,
            represented_pipeline.dep_structure_index,
        )

    def resolve_modes(self, graphene_info: ResolveInfo):
        represented_pipeline = self.get_represented_job()
        return [
            GrapheneMode(
                represented_pipeline.config_schema_snapshot.get_config_snap,
                self.resolve_id(graphene_info),
                mode_def_snap,
            )
            for mode_def_snap in sorted(
                represented_pipeline.mode_def_snaps, key=lambda item: item.name
            )
        ]

    def resolve_solid_handle(
        self, _graphene_info: ResolveInfo, handleID: str
    ) -> GrapheneSolidHandle | None:
        return build_solid_handles(self.get_represented_job()).get(handleID)

    def resolve_solid_handles(
        self, _graphene_info: ResolveInfo, parentHandleID: str | None = None
    ) -> Sequence[GrapheneSolidHandle]:
        handles = build_solid_handles(self.get_represented_job())

        if parentHandleID == "":
            handles = {key: handle for key, handle in handles.items() if not handle.parent}
        elif parentHandleID is not None:
            handles = {
                key: handle
                for key, handle in handles.items()
                if handle.parent and str(handle.parent.handleID) == parentHandleID
            }

        return [handles[key] for key in sorted(handles)]

    def resolve_tags(self, _graphene_info: ResolveInfo):
        represented_pipeline = self.get_represented_job()
        return [
            GraphenePipelineTag(key=key, value=value)
            for key, value in represented_pipeline.job_snapshot.tags.items()
            if get_tag_type(key) != TagType.HIDDEN
        ]

    def resolve_externalJobSource(self, graphene_info: ResolveInfo):
        """Retrieve the external job source from the job.

        The external job source comes from tags on the job. However, this resolver gets used in hot paths, so we only retrieve it if we can do so without extra queries.
        """
        return self.get_represented_job().get_external_job_source()

    def resolve_run_tags(self, _graphene_info: ResolveInfo):
        represented_pipeline = self.get_represented_job()
        return [
            GraphenePipelineTag(key=key, value=value)
            for key, value in (represented_pipeline.job_snapshot.run_tags or {}).items()
        ]

    def resolve_metadata_entries(self, _graphene_info: ResolveInfo) -> list[GrapheneMetadataEntry]:
        represented_pipeline = self.get_represented_job()
        return list(iterate_metadata_entries(represented_pipeline.job_snapshot.metadata))

    def resolve_solidSelection(self, _graphene_info: ResolveInfo):
        return self.get_represented_job().op_selection

    def resolve_runs(
        self, graphene_info: ResolveInfo, cursor: str | None = None, limit: int | None = None
    ) -> Sequence[GrapheneRun]:
        pipeline = self.get_represented_job()
        if isinstance(pipeline, RemoteJob):
            runs_filter = RunsFilter(
                job_name=pipeline.name,
                tags={
                    REPOSITORY_LABEL_TAG: (
                        pipeline.get_remote_origin().repository_origin.get_label()
                    )
                },
            )
        else:
            runs_filter = RunsFilter(job_name=pipeline.name)
        return get_runs(graphene_info, runs_filter, cursor, limit)

    def resolve_schedules(self, graphene_info: ResolveInfo):
        represented_pipeline = self.get_represented_job()
        if not isinstance(represented_pipeline, RemoteJob):
            # this is an historical pipeline snapshot, so there are not any associated running
            # schedules
            return []

        pipeline_selector = represented_pipeline.handle.to_selector()
        schedules = get_schedules_for_job(graphene_info, pipeline_selector)
        return schedules

    def resolve_sensors(self, graphene_info: ResolveInfo):
        represented_pipeline = self.get_represented_job()
        if not isinstance(represented_pipeline, RemoteJob):
            # this is an historical pipeline snapshot, so there are not any associated running
            # sensors
            return []

        pipeline_selector = represented_pipeline.handle.to_selector()
        sensors = get_sensors_for_job(graphene_info, pipeline_selector)
        return sensors

    def resolve_parent_snapshot_id(self, _graphene_info: ResolveInfo):
        lineage_snapshot = self.get_represented_job().job_snapshot.lineage_snapshot
        if lineage_snapshot:
            return lineage_snapshot.parent_snapshot_id
        else:
            return None

    def resolve_graph_name(self, _graphene_info: ResolveInfo):
        return self.get_represented_job().get_graph_name()


class GrapheneIPipelineSnapshot(graphene.Interface):
    name = graphene.NonNull(graphene.String)
    description = graphene.String()
    owners = non_null_list(GrapheneDefinitionOwner)
    pipeline_snapshot_id = graphene.NonNull(graphene.String)
    dagster_types = non_null_list(GrapheneDagsterType)
    dagster_type_or_error = graphene.Field(
        graphene.NonNull(GrapheneDagsterTypeOrError),
        dagsterTypeName=graphene.Argument(graphene.NonNull(graphene.String)),
    )
    solids = non_null_list(GrapheneSolid)
    modes = non_null_list(GrapheneMode)
    solid_handles = graphene.Field(
        non_null_list(GrapheneSolidHandle), parentHandleID=graphene.String()
    )
    solid_handle = graphene.Field(
        GrapheneSolidHandle,
        handleID=graphene.Argument(graphene.NonNull(graphene.String)),
    )
    tags = non_null_list(GraphenePipelineTag)
    metadata_entries = non_null_list(GrapheneMetadataEntry)
    runs = graphene.Field(
        non_null_list(GrapheneRun),
        cursor=graphene.String(),
        limit=graphene.Int(),
    )
    schedules = non_null_list(GrapheneSchedule)
    sensors = non_null_list(GrapheneSensor)
    parent_snapshot_id = graphene.String()
    graph_name = graphene.NonNull(graphene.String)

    class Meta:
        name = "IPipelineSnapshot"


class GraphenePipelinePreset(graphene.ObjectType):
    name = graphene.NonNull(graphene.String)
    solidSelection = graphene.List(graphene.NonNull(graphene.String))
    runConfigYaml = graphene.NonNull(graphene.String)
    mode = graphene.NonNull(graphene.String)
    tags = non_null_list(GraphenePipelineTag)

    class Meta:
        name = "PipelinePreset"

    def __init__(self, active_preset_data, pipeline_name):
        super().__init__()
        self._active_preset_data = check.inst_param(
            active_preset_data, "active_preset_data", PresetSnap
        )
        self._job_name = check.str_param(pipeline_name, "pipeline_name")

    def resolve_name(self, _graphene_info: ResolveInfo):
        return self._active_preset_data.name

    def resolve_solidSelection(self, _graphene_info: ResolveInfo):
        return self._active_preset_data.op_selection

    def resolve_runConfigYaml(self, _graphene_info: ResolveInfo):
        return dump_run_config_yaml(self._active_preset_data.run_config) or ""

    def resolve_mode(self, _graphene_info: ResolveInfo):
        return self._active_preset_data.mode

    def resolve_tags(self, _graphene_info: ResolveInfo):
        return [
            GraphenePipelineTag(key=key, value=value)
            for key, value in self._active_preset_data.tags.items()
        ]


class GraphenePipeline(GrapheneIPipelineSnapshotMixin, graphene.ObjectType):
    id = graphene.NonNull(graphene.ID)
    presets = non_null_list(GraphenePipelinePreset)
    isJob = graphene.NonNull(graphene.Boolean)
    isAssetJob = graphene.NonNull(graphene.Boolean)
    repository = graphene.NonNull("dagster_graphql.schema.external.GrapheneRepository")
    partitionKeysOrError = graphene.Field(
        graphene.NonNull(GraphenePartitionKeys),
        cursor=graphene.String(),
        limit=graphene.Int(),
        reverse=graphene.Boolean(),
        selected_asset_keys=graphene.Argument(
            graphene.List(graphene.NonNull(GrapheneAssetKeyInput))
        ),
    )
    partition = graphene.Field(
        "dagster_graphql.schema.partition_sets.GrapheneJobSelectionPartition",
        partition_name=graphene.NonNull(graphene.String),
        selected_asset_keys=graphene.Argument(
            graphene.List(graphene.NonNull(GrapheneAssetKeyInput))
        ),
    )
    hasLaunchExecutionPermission = graphene.NonNull(graphene.Boolean)
    hasLaunchReexecutionPermission = graphene.NonNull(graphene.Boolean)
    nodeNames = non_null_list(graphene.String)

    class Meta:  # pyright: ignore[reportIncompatibleVariableOverride]
        interfaces = (GrapheneSolidContainer, GrapheneIPipelineSnapshot)
        name = "Pipeline"

    def __init__(
        self, remote_job: RemoteJob, batch_loader: RepositoryScopedBatchLoader | None = None
    ):
        super().__init__()
        self._remote_job = check.inst_param(remote_job, "remote_job", RemoteJob)
        self._batch_loader = check.opt_inst_param(
            batch_loader, "batch_loader", RepositoryScopedBatchLoader
        )

    def resolve_id(self, _graphene_info: ResolveInfo):
        return self._remote_job.get_remote_origin_id()

    def get_represented_job(self) -> RepresentedJob:
        return self._remote_job

    def resolve_nodeNames(self, _graphene_info: ResolveInfo):
        return self._remote_job.node_names

    def resolve_presets(self, _graphene_info: ResolveInfo):
        return [
            GraphenePipelinePreset(preset, self._remote_job.name)
            for preset in sorted(self._remote_job.active_presets, key=lambda item: item.name)
        ]

    def resolve_isJob(self, _graphene_info: ResolveInfo):
        return True

    def resolve_isAssetJob(self, graphene_info: ResolveInfo):
        if is_implicit_asset_job_name(self._remote_job.name):
            return True

        if self._batch_loader:
            return bool(self._batch_loader.repository.get_asset_keys_in_job(self._remote_job.name))

        return bool(
            graphene_info.context.get_asset_keys_in_job(
                self._remote_job.handle.to_selector(),
            )
        )

    def resolve_repository(self, graphene_info: ResolveInfo):
        from dagster_graphql.schema.external import GrapheneRepository

        return GrapheneRepository(self._remote_job.repository_handle)

    def resolve_partitionKeysOrError(
        self,
        graphene_info: ResolveInfo,
        cursor: str | None = None,
        limit: int | None = None,
        reverse: bool | None = None,
        selected_asset_keys: list[GrapheneAssetKeyInput] | None = None,
    ) -> GraphenePartitionKeys:
        result = graphene_info.context.get_partition_names(
            repository_selector=self._remote_job.repository_handle.to_selector(),
            job_name=self._remote_job.name,
            selected_asset_keys=_asset_key_input_list_to_asset_key_set(selected_asset_keys),
            instance=graphene_info.context.instance,
        )

        if isinstance(result, PartitionExecutionErrorSnap):
            raise DagsterUserCodeProcessError.from_error_info(result.error)

        all_partition_keys = result.partition_names

        return GraphenePartitionKeys(
            partitionKeys=apply_cursor_limit_reverse(
                all_partition_keys, cursor=cursor, limit=limit, reverse=reverse or False
            )
        )

    def resolve_partition(
        self,
        graphene_info: ResolveInfo,
        partition_name: str,
        selected_asset_keys: list[GrapheneAssetKeyInput] | None = None,
    ) -> "GrapheneJobSelectionPartition":
        from dagster_graphql.schema.partition_sets import GrapheneJobSelectionPartition

        return GrapheneJobSelectionPartition(
            remote_job=self._remote_job,
            partition_name=partition_name,
            selected_asset_keys=_asset_key_input_list_to_asset_key_set(selected_asset_keys),
        )

    def resolve_hasLaunchExecutionPermission(self, graphene_info: ResolveInfo) -> bool:
        return has_permission_for_definition(
            graphene_info, Permissions.LAUNCH_PIPELINE_EXECUTION, self._remote_job
        )

    def resolve_hasLaunchReexecutionPermission(self, graphene_info: ResolveInfo) -> bool:
        return has_permission_for_definition(
            graphene_info, Permissions.LAUNCH_PIPELINE_REEXECUTION, self._remote_job
        )


class GrapheneJob(GraphenePipeline):
    class Meta:  # pyright: ignore[reportIncompatibleVariableOverride]
        interfaces = (GrapheneSolidContainer, GrapheneIPipelineSnapshot)
        name = "Job"

    # doesn't inherit from base class
    def __init__(
        self, remote_job: RemoteJob, batch_loader: RepositoryScopedBatchLoader | None = None
    ):
        super().__init__()  # pyright: ignore[reportCallIssue]
        self._remote_job = check.inst_param(remote_job, "remote_job", RemoteJob)
        self._batch_loader = check.opt_inst_param(
            batch_loader, "batch_loader", RepositoryScopedBatchLoader
        )


class GrapheneGraph(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneSolidContainer,)
        name = "Graph"

    id = graphene.NonNull(graphene.ID)
    name = graphene.NonNull(graphene.String)
    description = graphene.String()
    solid_handle = graphene.Field(
        GrapheneSolidHandle,
        handleID=graphene.Argument(graphene.NonNull(graphene.String)),
    )
    solid_handles = graphene.Field(
        non_null_list(GrapheneSolidHandle), parentHandleID=graphene.String()
    )
    modes = non_null_list(GrapheneMode)

    def __init__(self, remote_job: RemoteJob, solid_handle_id=None):
        self._remote_job = check.inst_param(remote_job, "remote_job", RemoteJob)
        self._solid_handle_id = check.opt_str_param(solid_handle_id, "solid_handle_id")
        super().__init__()

    def resolve_id(self, _graphene_info: ResolveInfo):
        if self._solid_handle_id:
            return f"{self._remote_job.get_remote_origin_id()}:solid:{self._solid_handle_id}"
        return f"graph:{self._remote_job.get_remote_origin_id()}"

    def resolve_name(self, _graphene_info: ResolveInfo):
        return self._remote_job.get_graph_name()

    def resolve_description(self, _graphene_info: ResolveInfo):
        return self._remote_job.description

    def resolve_solid_handle(
        self, _graphene_info: ResolveInfo, handleID: str
    ) -> GrapheneSolidHandle | None:
        return build_solid_handles(self._remote_job).get(handleID)

    def resolve_solid_handles(
        self, _graphene_info: ResolveInfo, parentHandleID: str | None = None
    ) -> Sequence[GrapheneSolidHandle]:
        handles = build_solid_handles(self._remote_job)

        if parentHandleID == "":
            handles = {key: handle for key, handle in handles.items() if not handle.parent}
        elif parentHandleID is not None:
            handles = {
                key: handle
                for key, handle in handles.items()
                if handle.parent and str(handle.parent.handleID) == parentHandleID
            }

        return [handles[key] for key in sorted(handles)]

    def resolve_modes(self, _graphene_info: ResolveInfo):
        # returns empty list... graphs don't have modes, this is a vestige of the old
        # pipeline explorer, which expected all solid containers to be pipelines
        return []


class GrapheneRunOrError(graphene.Union):
    class Meta:
        types = (GrapheneRun, GrapheneRunNotFoundError, GraphenePythonError)
        name = "RunOrError"


def _asset_key_input_list_to_asset_key_set(
    asset_keys: list[GrapheneAssetKeyInput] | None,
) -> AbstractSet[AssetKey] | None:
    return (
        {key_input.to_asset_key() for key_input in asset_keys} if asset_keys is not None else None
    )


class GrapheneAssetRecord(graphene.ObjectType):
    id = graphene.NonNull(graphene.String)
    key = graphene.NonNull(GrapheneAssetKey)

    class Meta:
        name = "AssetRecord"



from collections.abc import Sequence
from enum import Enum
from typing import TYPE_CHECKING, Optional, TypeAlias, Union, cast

import dagster._check as check
import graphene
from dagster import EventLogEntry
from dagster._core.definitions.asset_checks.asset_check_evaluation import (
    AssetCheckEvaluation,
    AssetCheckEvaluationTargetMaterializationData,
)
from dagster._core.definitions.asset_checks.asset_check_spec import AssetCheckSeverity
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.assets.graph.remote_asset_graph import RemoteAssetCheckNode
from dagster._core.definitions.declarative_automation.serialized_objects import (
    AutomationConditionSnapshot,
)
from dagster._core.definitions.partitions.snap import MultiPartitionsSnap
from dagster._core.events import DagsterEventType
from dagster._core.storage.asset_check_execution_record import (
    AssetCheckExecutionRecord,
    AssetCheckExecutionResolvedStatus,
)
from dagster._core.storage.dagster_run import RunRecord

from dagster_graphql.implementation.events import iterate_metadata_entries
from dagster_graphql.schema.auto_materialize_policy import GrapheneAutoMaterializePolicy
from dagster_graphql.schema.automation_condition import GrapheneAutomationCondition
from dagster_graphql.schema.entity_key import GrapheneAssetKey
from dagster_graphql.schema.errors import GrapheneError
from dagster_graphql.schema.metadata import GrapheneMetadataEntry
from dagster_graphql.schema.util import ResolveInfo, non_null_list

if TYPE_CHECKING:
    from dagster_graphql.schema.partition_sets import (
        GrapheneDimensionPartitionKeys,
        GraphenePartitionDefinition,
    )

GrapheneAssetCheckExecutionResolvedStatus = graphene.Enum.from_enum(
    AssetCheckExecutionResolvedStatus
)


class GrapheneAssetCheckEvaluationTargetMaterializationData(graphene.ObjectType):
    storageId = graphene.NonNull(graphene.ID)
    runId = graphene.NonNull(graphene.String)
    timestamp = graphene.NonNull(graphene.Float)

    class Meta:
        name = "AssetCheckEvaluationTargetMaterializationData"

    def __init__(self, target_materialization_data: AssetCheckEvaluationTargetMaterializationData):
        self.storageId = target_materialization_data.storage_id
        self.runId = target_materialization_data.run_id
        self.timestamp = target_materialization_data.timestamp


GrapheneAssetCheckSeverity = graphene.Enum.from_enum(AssetCheckSeverity)


class GrapheneAssetCheckEvaluation(graphene.ObjectType):
    timestamp = graphene.Field(
        graphene.NonNull(graphene.Float), description="When the check evaluation was stored"
    )
    checkName = graphene.NonNull(graphene.String)
    assetKey = graphene.NonNull(GrapheneAssetKey)
    targetMaterialization = graphene.Field(GrapheneAssetCheckEvaluationTargetMaterializationData)
    metadataEntries = non_null_list(GrapheneMetadataEntry)
    severity = graphene.NonNull(GrapheneAssetCheckSeverity)
    description = graphene.String()
    partition = graphene.String()

    # NOTE: this should be renamed passed
    success = graphene.NonNull(graphene.Boolean)

    class Meta:
        name = "AssetCheckEvaluation"

    def __init__(self, evaluation_event: EventLogEntry):
        self.timestamp = evaluation_event.timestamp

        evaluation_data = cast(
            "AssetCheckEvaluation",
            check.not_none(evaluation_event.dagster_event).event_specific_data,
        )
        target_materialization_data = evaluation_data.target_materialization_data
        self.targetMaterialization = (
            GrapheneAssetCheckEvaluationTargetMaterializationData(target_materialization_data)
            if target_materialization_data
            else None
        )

        self.metadataEntries = list(iterate_metadata_entries(evaluation_data.metadata))
        self.severity = evaluation_data.severity
        self.success = evaluation_data.passed
        self.checkName = evaluation_data.check_name
        self.assetKey = evaluation_data.asset_key
        self.description = evaluation_data.description
        self.partition = evaluation_data.partition


class GrapheneAssetCheckExecution(graphene.ObjectType):
    id = graphene.NonNull(graphene.String)
    runId = graphene.NonNull(graphene.String)
    status = graphene.NonNull(GrapheneAssetCheckExecutionResolvedStatus)
    evaluation = graphene.Field(GrapheneAssetCheckEvaluation)
    timestamp = graphene.Field(
        graphene.NonNull(graphene.Float), description="When the check run started"
    )
    partition = graphene.Field(graphene.String)
    stepKey = graphene.Field(graphene.String)
    run = graphene.Field("dagster_graphql.schema.pipelines.pipeline.GrapheneRun")

    class Meta:
        name = "AssetCheckExecution"

    def __init__(self, execution: AssetCheckExecutionRecord):
        super().__init__()
        self._execution = execution
        self.id = str(execution.id)
        self.runId = execution.run_id
        self.evaluation = (
            GrapheneAssetCheckEvaluation(execution.event)
            if execution.event
            and execution.event.dagster_event_type == DagsterEventType.ASSET_CHECK_EVALUATION
            else None
        )
        self.timestamp = execution.create_timestamp
        self.stepKey = execution.event.step_key if execution.event else None
        self.partition = execution.partition

    async def resolve_status(
        self, graphene_info: "ResolveInfo"
    ) -> AssetCheckExecutionResolvedStatus:
        return await self._execution.resolve_status(graphene_info.context)

    async def resolve_run(self, graphene_info: "ResolveInfo"):
        from dagster_graphql.schema.pipelines.pipeline import GrapheneRun

        if self.runId:
            run_record = await RunRecord.gen(graphene_info.context, self.runId)
        else:
            run_record = None

        return GrapheneRun(run_record) if run_record else None


class GrapheneAssetCheckCanExecuteIndividually(graphene.Enum):
    class Meta:
        name = "AssetCheckCanExecuteIndividually"

    CAN_EXECUTE = "CAN_EXECUTE"
    REQUIRES_MATERIALIZATION = "REQUIRES_MATERIALIZATION"
    NEEDS_USER_CODE_UPGRADE = "NEEDS_USER_CODE_UPGRADE"


# Enum for asset check partition range status
class AssetCheckPartitionRangeStatus(Enum):
    """Status for asset check partition ranges."""

    SUCCEEDED = "SUCCEEDED"
    FAILED = "FAILED"
    IN_PROGRESS = "IN_PROGRESS"
    SKIPPED = "SKIPPED"
    EXECUTION_FAILED = "EXECUTION_FAILED"


GrapheneAssetCheckPartitionRangeStatus = graphene.Enum.from_enum(AssetCheckPartitionRangeStatus)


# Time partition types
class GrapheneAssetCheckTimePartitionRangeStatus(graphene.ObjectType):
    """Time range with associated check status.

    Uses the same fields as GrapheneTimePartitionRange but with check-specific status.
    """

    startTime = graphene.NonNull(graphene.Float)
    endTime = graphene.NonNull(graphene.Float)
    startKey = graphene.NonNull(graphene.String)
    endKey = graphene.NonNull(graphene.String)
    status = graphene.NonNull(GrapheneAssetCheckPartitionRangeStatus)

    class Meta:
        name = "AssetCheckTimePartitionRangeStatus"


class GrapheneAssetCheckTimePartitionStatuses(graphene.ObjectType):
    """Time-partitioned asset check statuses with range compression.

    Each range represents a contiguous set of partitions with the same status.
    Unlike assets, asset checks expose all 6 statuses as separate ranges.
    """

    ranges = non_null_list(GrapheneAssetCheckTimePartitionRangeStatus)

    class Meta:
        name = "AssetCheckTimePartitionStatuses"


# Default partition type
class GrapheneAssetCheckDefaultPartitionStatuses(graphene.ObjectType):
    """Static-partitioned asset check statuses as partition key lists.

    Each field contains the list of partition keys with that status.
    """

    succeededPartitions = non_null_list(graphene.String)
    failedPartitions = non_null_list(graphene.String)
    inProgressPartitions = non_null_list(graphene.String)
    skippedPartitions = non_null_list(graphene.String)
    executionFailedPartitions = non_null_list(graphene.String)

    class Meta:
        name = "AssetCheckDefaultPartitionStatuses"


# Multi-partition types
class GrapheneAssetCheckPartitionStatus1D(graphene.Union):
    """Union for secondary dimension (can be Time or Default)."""

    class Meta:
        types = (
            GrapheneAssetCheckTimePartitionStatuses,
            GrapheneAssetCheckDefaultPartitionStatuses,
        )
        name = "AssetCheckPartitionStatus1D"


class GrapheneAssetCheckMultiPartitionRangeStatuses(graphene.ObjectType):
    """Range in primary dimension with secondary dimension statuses.

    The primary dimension is compressed using run-length encoding (consecutive primary
    partitions with identical secondary dimension status patterns are grouped).
    """

    primaryDimStartKey = graphene.NonNull(graphene.String)
    primaryDimEndKey = graphene.NonNull(graphene.String)
    primaryDimStartTime = graphene.Field(graphene.Float)
    primaryDimEndTime = graphene.Field(graphene.Float)
    secondaryDim = graphene.NonNull(GrapheneAssetCheckPartitionStatus1D)

    class Meta:
        name = "AssetCheckMultiPartitionRangeStatuses"


class GrapheneAssetCheckMultiPartitionStatuses(graphene.ObjectType):
    """Multi-partitioned asset check statuses with 2D run-length encoding."""

    ranges = non_null_list(GrapheneAssetCheckMultiPartitionRangeStatuses)
    primaryDimensionName = graphene.NonNull(graphene.String)

    class Meta:
        name = "AssetCheckMultiPartitionStatuses"


# Union type (replaces old ObjectType)
class GrapheneAssetCheckPartitionStatuses(graphene.Union):
    """Union type for asset check partition statuses.

    The concrete type returned depends on the partitions definition:
    - TimeWindowPartitionsDefinition -> AssetCheckTimePartitionStatuses
    - MultiPartitionsDefinition -> AssetCheckMultiPartitionStatuses
    - Other (Static, Dynamic) -> AssetCheckDefaultPartitionStatuses
    """

    class Meta:
        types = (
            GrapheneAssetCheckDefaultPartitionStatuses,
            GrapheneAssetCheckMultiPartitionStatuses,
            GrapheneAssetCheckTimePartitionStatuses,
        )
        name = "AssetCheckPartitionStatuses"


class GrapheneAssetCheck(graphene.ObjectType):
    name = graphene.NonNull(graphene.String)
    assetKey = graphene.NonNull(GrapheneAssetKey)
    description = graphene.String()
    jobNames = non_null_list(graphene.String)
    executionForLatestMaterialization = graphene.Field(GrapheneAssetCheckExecution)
    canExecuteIndividually = graphene.NonNull(GrapheneAssetCheckCanExecuteIndividually)
    blocking = graphene.NonNull(graphene.Boolean)
    additionalAssetKeys = non_null_list(GrapheneAssetKey)
    automationCondition = graphene.Field(GrapheneAutomationCondition)
    partitionDefinition = graphene.Field(
        "dagster_graphql.schema.partition_sets.GraphenePartitionDefinition"
    )
    partitionKeysByDimension = graphene.Field(
        non_null_list("dagster_graphql.schema.partition_sets.GrapheneDimensionPartitionKeys"),
        startIdx=graphene.Int(),
        endIdx=graphene.Int(),
    )
    partitionStatuses = graphene.Field(GrapheneAssetCheckPartitionStatuses)

    class Meta:
        name = "AssetCheck"

    def __init__(
        self,
        remote_node: RemoteAssetCheckNode,
    ):
        self._remote_node = remote_node
        self._asset_check = remote_node.asset_check

    def resolve_assetKey(self, _) -> AssetKey:
        return self._asset_check.asset_key

    def resolve_name(self, _) -> str:
        return self._asset_check.name

    def resolve_description(self, _) -> str | None:
        return self._asset_check.description

    def resolve_jobNames(self, _) -> Sequence[str]:
        return self._asset_check.job_names

    async def resolve_executionForLatestMaterialization(
        self, graphene_info: ResolveInfo
    ) -> GrapheneAssetCheckExecution | None:
        record = await AssetCheckExecutionRecord.gen(graphene_info.context, self._asset_check.key)
        return (
            GrapheneAssetCheckExecution(record)
            if record and await record.targets_latest_materialization(graphene_info.context)
            else None
        )

    def resolve_canExecuteIndividually(self, _: ResolveInfo):
        return (
            GrapheneAssetCheckCanExecuteIndividually.CAN_EXECUTE
            if len(self._remote_node.execution_set_entity_keys) <= 1
            # NOTE: once we support multi checks, we'll need to add a case for
            # non subsettable multi checks
            else GrapheneAssetCheckCanExecuteIndividually.REQUIRES_MATERIALIZATION
        )

    def resolve_blocking(self, _) -> bool:
        return self._asset_check.blocking

    def resolve_additionalAssetKeys(self, _) -> Sequence[GrapheneAssetKey]:
        return [
            GrapheneAssetKey(path=asset_key.path)
            for asset_key in self._asset_check.additional_asset_keys
        ]

    def resolve_automationCondition(
        self, _graphene_info: ResolveInfo
    ) -> GrapheneAutoMaterializePolicy | None:
        automation_condition = (
            self._asset_check.automation_condition_snapshot
            or self._asset_check.automation_condition
        )
        if automation_condition:
            return GrapheneAutomationCondition(
                # we only store one of automation_condition or automation_condition_snapshot
                automation_condition
                if isinstance(automation_condition, AutomationConditionSnapshot)
                else automation_condition.get_snapshot()
            )
        return None

    def resolve_partitionDefinition(
        self, _graphene_info: ResolveInfo
    ) -> Optional["GraphenePartitionDefinition"]:
        from dagster_graphql.schema.partition_sets import GraphenePartitionDefinition

        partitions_snap = self._asset_check.partitions_def_snapshot
        if partitions_snap:
            return GraphenePartitionDefinition(partitions_snap)
        return None

    def resolve_partitionKeysByDimension(
        self,
        graphene_info: ResolveInfo,
        startIdx: int | None = None,
        endIdx: int | None = None,
    ) -> Sequence["GrapheneDimensionPartitionKeys"]:
        from dagster_graphql.schema.partition_sets import (
            GrapheneDimensionPartitionKeys,
            GraphenePartitionDefinitionType,
            get_partition_keys_from_snap,
        )

        # Accepts startIdx and endIdx arguments. This will be used to select a range of
        # time partitions. StartIdx is inclusive, endIdx is exclusive.
        # For non time partition definitions, these arguments will be ignored
        # and the full list of partition keys will be returned.
        if not self._asset_check.partitions_def_snapshot:
            return []

        dynamic_partitions_loader = graphene_info.context.dynamic_partitions_loader
        if isinstance(self._asset_check.partitions_def_snapshot, MultiPartitionsSnap):
            return [
                GrapheneDimensionPartitionKeys(
                    name=dimension.name,
                    partition_keys=get_partition_keys_from_snap(
                        dimension.partitions,
                        dynamic_partitions_loader,
                        startIdx,
                        endIdx,
                    ),
                    type=GraphenePartitionDefinitionType.from_partition_def_data(
                        dimension.partitions
                    ),
                )
                for dimension in cast(
                    "MultiPartitionsSnap",
                    self._asset_check.partitions_def_snapshot,
                ).partition_dimensions
            ]

        return [
            GrapheneDimensionPartitionKeys(
                name="default",
                type=GraphenePartitionDefinitionType.from_partition_def_data(
                    self._asset_check.partitions_def_snapshot
                ),
                partition_keys=get_partition_keys_from_snap(
                    partitions_snap=self._asset_check.partitions_def_snapshot,
                    dynamic_partitions_loader=dynamic_partitions_loader,
                    start_idx=startIdx,
                    end_idx=endIdx,
                ),
            )
        ]

    async def resolve_partitionStatuses(
        self, graphene_info: ResolveInfo
    ) -> (
        Union[
            "GrapheneAssetCheckTimePartitionStatuses",
            "GrapheneAssetCheckDefaultPartitionStatuses",
            "GrapheneAssetCheckMultiPartitionStatuses",
        ]
        | None
    ):
        """Resolve partition statuses using union type with efficient representation."""
        from dagster._core.definitions.asset_checks.asset_check_spec import AssetCheckKey
        from dagster._core.storage.asset_check_state import AssetCheckState

        from dagster_graphql.implementation.fetch_asset_checks import (
            build_asset_check_partition_statuses,
        )

        # Only return partition statuses for partitioned checks
        if not self._asset_check.partitions_def_snapshot:
            return None

        check_key = AssetCheckKey(
            asset_key=self._asset_check.asset_key, name=self._asset_check.name
        )

        # Get current partition definition
        current_partition_def = (
            self._asset_check.partitions_def_snapshot.get_partitions_definition()
        )

        # Get partition status cache using the cache service (handles storage + reconciliation)
        partition_status = await AssetCheckState.gen(
            graphene_info.context, (check_key, current_partition_def)
        )
        if not partition_status:
            return None

        # Use builder function to construct appropriate union type
        return build_asset_check_partition_statuses(
            key=check_key,
            asset_check_state=partition_status,
            partitions_def=current_partition_def,
            dynamic_partitions_store=graphene_info.context.instance,
        )


class GrapheneAssetChecks(graphene.ObjectType):
    checks = non_null_list(GrapheneAssetCheck)

    class Meta:
        name = "AssetChecks"


class GrapheneAssetCheckNeedsMigrationError(graphene.ObjectType):
    message = graphene.NonNull(graphene.String)

    class Meta:
        interfaces = (GrapheneError,)
        name = "AssetCheckNeedsMigrationError"


class GrapheneAssetCheckNeedsAgentUpgradeError(graphene.ObjectType):
    message = graphene.NonNull(graphene.String)

    class Meta:
        interfaces = (GrapheneError,)
        name = "AssetCheckNeedsAgentUpgradeError"


class GrapheneAssetCheckNeedsUserCodeUpgrade(graphene.ObjectType):
    message = graphene.NonNull(graphene.String)

    class Meta:
        interfaces = (GrapheneError,)
        name = "AssetCheckNeedsUserCodeUpgrade"


AssetChecksOrErrorUnion: TypeAlias = (
    GrapheneAssetCheckNeedsMigrationError
    | GrapheneAssetCheckNeedsUserCodeUpgrade
    | GrapheneAssetCheckNeedsAgentUpgradeError
    | GrapheneAssetChecks
)


class GrapheneAssetChecksOrError(graphene.Union):
    class Meta:
        types = (
            GrapheneAssetChecks,
            GrapheneAssetCheckNeedsMigrationError,
            GrapheneAssetCheckNeedsUserCodeUpgrade,
            GrapheneAssetCheckNeedsAgentUpgradeError,
        )
        name = "AssetChecksOrError"


class GrapheneAssetCheckNotFoundError(graphene.ObjectType):
    message = graphene.NonNull(graphene.String)

    class Meta:
        interfaces = (GrapheneError,)
        name = "AssetCheckNotFoundError"


class GrapheneAssetCheckOrError(graphene.Union):
    class Meta:
        types = (
            GrapheneAssetCheck,
            GrapheneAssetCheckNotFoundError,
            GrapheneAssetCheckNeedsMigrationError,
            GrapheneAssetCheckNeedsUserCodeUpgrade,
            GrapheneAssetCheckNeedsAgentUpgradeError,
        )
        name = "AssetCheckOrError"



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, cast

import dagster._check as check
from dagster._core.definitions.selector import JobSubsetSelector
from dagster._core.execution.plan.resume_retry import ReexecutionStrategy
from dagster._core.storage.dagster_run import DagsterRun, RunsFilter
from dagster._core.workspace.permissions import Permissions

from dagster_graphql.implementation.execution.run_lifecycle import create_valid_pipeline_run
from dagster_graphql.implementation.external import get_remote_job_or_raise
from dagster_graphql.implementation.utils import (
    ExecutionMetadata,
    ExecutionParams,
    assert_permission_for_run,
)

if TYPE_CHECKING:
    from dagster._core.instance import DagsterInstance

    from dagster_graphql.schema.runs import GrapheneLaunchRunSuccess
    from dagster_graphql.schema.util import ResolveInfo


async def launch_pipeline_reexecution(
    graphene_info: "ResolveInfo", execution_params: ExecutionParams
) -> "GrapheneLaunchRunSuccess":
    return await _launch_pipeline_execution(graphene_info, execution_params, is_reexecuted=True)


async def launch_pipeline_execution(
    graphene_info: "ResolveInfo", execution_params: ExecutionParams
) -> "GrapheneLaunchRunSuccess":
    return await _launch_pipeline_execution(graphene_info, execution_params)


async def do_launch(
    graphene_info: "ResolveInfo", execution_params: ExecutionParams, is_reexecuted: bool = False
) -> DagsterRun:
    check.inst_param(execution_params, "execution_params", ExecutionParams)
    check.bool_param(is_reexecuted, "is_reexecuted")

    if is_reexecuted:
        # required fields for re-execution
        execution_metadata = check.inst_param(
            execution_params.execution_metadata, "execution_metadata", ExecutionMetadata
        )
        check.str_param(execution_metadata.root_run_id, "root_run_id")
        check.str_param(execution_metadata.parent_run_id, "parent_run_id")
    remote_job = await get_remote_job_or_raise(graphene_info, execution_params.selector)
    dagster_run = create_valid_pipeline_run(graphene_info.context, remote_job, execution_params)

    return graphene_info.context.instance.submit_run(
        dagster_run.run_id,
        workspace=graphene_info.context,
    )


async def _launch_pipeline_execution(
    graphene_info: "ResolveInfo", execution_params: ExecutionParams, is_reexecuted: bool = False
) -> "GrapheneLaunchRunSuccess":
    from dagster_graphql.schema.pipelines.pipeline import GrapheneRun
    from dagster_graphql.schema.runs import GrapheneLaunchRunSuccess

    check.inst_param(execution_params, "execution_params", ExecutionParams)
    check.bool_param(is_reexecuted, "is_reexecuted")

    run = await do_launch(graphene_info, execution_params, is_reexecuted)
    records = graphene_info.context.instance.get_run_records(RunsFilter(run_ids=[run.run_id]))

    return GrapheneLaunchRunSuccess(run=GrapheneRun(records[0]))


async def launch_reexecution_from_parent_run(
    graphene_info: "ResolveInfo",
    parent_run_id: str,
    strategy: str,
    extra_tags: Mapping[str, Any] | None = None,
    use_parent_run_tags: bool | None = None,
) -> "GrapheneLaunchRunSuccess":
    """Launch a re-execution by referencing the parent run id."""
    from dagster_graphql.schema.pipelines.pipeline import GrapheneRun
    from dagster_graphql.schema.runs import GrapheneLaunchRunSuccess

    check.str_param(parent_run_id, "parent_run_id")

    instance: DagsterInstance = graphene_info.context.instance
    parent_run = check.not_none(
        instance.get_run_by_id(parent_run_id), f"Could not find parent run with id: {parent_run_id}"
    )
    origin = check.not_none(parent_run.remote_job_origin)
    selector = JobSubsetSelector(
        location_name=origin.repository_origin.code_location_origin.location_name,
        repository_name=origin.repository_origin.repository_name,
        job_name=parent_run.job_name,
        asset_selection=parent_run.asset_selection,
        asset_check_selection=parent_run.asset_check_selection,
        op_selection=None,
    )

    repo_location = graphene_info.context.get_code_location(selector.location_name)
    external_pipeline = await get_remote_job_or_raise(graphene_info, selector)
    assert_permission_for_run(
        graphene_info,
        Permissions.LAUNCH_PIPELINE_REEXECUTION,
        parent_run,
    )

    run = instance.create_reexecuted_run(
        parent_run=cast("DagsterRun", parent_run),
        request_context=graphene_info.context,
        code_location=repo_location,
        remote_job=external_pipeline,
        strategy=ReexecutionStrategy(strategy),
        extra_tags=extra_tags,
        use_parent_run_tags=use_parent_run_tags
        if use_parent_run_tags is not None
        else True,  # inherit whatever tags were set on the parent run at launch time
    )
    graphene_info.context.instance.submit_run(
        run.run_id,
        workspace=graphene_info.context,
    )

    # return run with updateTime
    records = graphene_info.context.instance.get_run_records(RunsFilter(run_ids=[run.run_id]))
    return GrapheneLaunchRunSuccess(run=GrapheneRun(records[0]))
